/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.gui.utils;

import com.net2plan.gui.utils.topology.INetworkCallback;
import com.net2plan.internal.ErrorHandling;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.internal.Constants.NetworkElementType;
import com.net2plan.utils.Pair;
import java.awt.BorderLayout;
import java.awt.Dialog;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.table.DefaultTableModel;

/**
 * Utility to edit attributes from a set of elements (i.e. nodes, links, and so on).
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.3.0
 */
public class AttributeEditor extends JDialog implements ActionListener
{
	private final JTable table;
	private final Object itemIds;
	
	/**
	 * Default constructor.
	 *
	 * @param callback Reference to the handler of the network design
	 * @param type Type of element (i.e. layers, nodes, links, and so on)
	 * @since 0.3.0
	 */
	public AttributeEditor(final INetworkCallback callback, final NetworkElementType type)
	{
		NetPlan netPlan = callback.getDesign();
		
		Object[][] data;
		String[] columnNames;
		final Set<String> attributes = new TreeSet<String>();
		String dialogHeader;

		switch (type)
		{
			case LAYER:
				long[] layerIds = netPlan.getLayerIdsVector();
				itemIds = layerIds;
				int L = layerIds.length;

				columnNames = new String[L + 1];
				columnNames[0] = "Attribute";
				for (int l = 0; l < L; l++)
				{
					long layerId = layerIds[l];
					columnNames[l + 1] = generateColumnName(type, layerId);
					attributes.addAll(netPlan.getLayerAttributeMap(layerId).keySet());
				}

				if (attributes.isEmpty())
				{
					data = new Object[1][L + 1];
				}
				else
				{
					data = new Object[attributes.size()][L + 1];

					int itemId = 0;
					Iterator<String> it = attributes.iterator();
					while (it.hasNext())
					{
						String attribute = it.next();
						data[itemId][0] = attribute;
						for (int l = 0; l < L; l++)
						{
							long layerId = layerIds[l];
							data[itemId][l + 1] = netPlan.getLayerAttribute(layerId, attribute);
							if (data[itemId][l + 1] == null)
							{
								data[itemId][l + 1] = "";
							}
						}

						itemId++;
					}
				}

				dialogHeader = "Edit layer attributes";

				break;

			case NODE:
				long[] nodeIds = netPlan.getNodeIdsVector();
				itemIds = nodeIds;
				int N = nodeIds.length;

				columnNames = new String[N + 1];
				columnNames[0] = "Attribute";
				for (int n = 0; n < N; n++)
				{
					long nodeId = nodeIds[n];
					columnNames[n + 1] = generateColumnName(type, nodeId);
					attributes.addAll(netPlan.getNodeAttributeMap(nodeId).keySet());
				}

				if (attributes.isEmpty())
				{
					data = new Object[1][N + 1];
				}
				else
				{
					data = new Object[attributes.size()][N + 1];

					int itemId = 0;
					Iterator<String> it = attributes.iterator();
					while (it.hasNext())
					{
						String attribute = it.next();
						data[itemId][0] = attribute;
						for (int n = 0; n < N; n++)
						{
							long nodeId = nodeIds[n];
							data[itemId][n + 1] = netPlan.getNodeAttribute(nodeId, attribute);
							if (data[itemId][n + 1] == null)
							{
								data[itemId][n + 1] = "";
							}
						}

						itemId++;
					}
				}

				dialogHeader = "Edit node attributes";

				break;

			case LINK:
				long[] linkIds = netPlan.getLinkIdsVector();
				itemIds = linkIds;
				int E = linkIds.length;

				columnNames = new String[E + 1];
				columnNames[0] = "Attribute";
				for (int e = 0; e < E; e++)
				{
					long linkId = linkIds[e];
					columnNames[e + 1] = generateColumnName(type, linkId);
					attributes.addAll(netPlan.getLinkAttributeMap(linkId).keySet());
				}

				if (attributes.isEmpty())
				{
					data = new Object[1][E + 1];
				}
				else
				{
					data = new Object[attributes.size()][E + 1];

					int itemId = 0;
					Iterator<String> it = attributes.iterator();
					while (it.hasNext())
					{
						String attribute = it.next();
						data[itemId][0] = attribute;
						for (int e = 0; e < E; e++)
						{
							long linkId = linkIds[e];
							data[itemId][e + 1] = netPlan.getLinkAttribute(linkId, attribute);
							if (data[itemId][e + 1] == null)
							{
								data[itemId][e + 1] = "";
							}
						}

						itemId++;
					}
				}

				dialogHeader = "Edit link attributes";
				break;

			case DEMAND:
				long[] demandIds = netPlan.getDemandIdsVector();
				itemIds = demandIds;
				int D = demandIds.length;

				columnNames = new String[D + 1];
				columnNames[0] = "Attribute";
				for (int d = 0; d < D; d++)
				{
					long demandId = demandIds[d];
					columnNames[d] = generateColumnName(type, demandId);
					attributes.addAll(netPlan.getDemandAttributeMap(demandId).keySet());
				}

				if (attributes.isEmpty())
				{
					data = new Object[1][D + 1];
				}
				else
				{
					data = new Object[attributes.size()][D + 1];

					int itemId = 0;
					Iterator<String> it = attributes.iterator();
					while (it.hasNext())
					{
						String attribute = it.next();
						data[itemId][0] = attribute;
						for (int d = 0; d < D; d++)
						{
							long demandId = demandIds[d];
							data[itemId][d + 1] = netPlan.getDemandAttribute(demandId, attribute);
							if (data[itemId][d + 1] == null)
							{
								data[itemId][d + 1] = "";
							}
						}

						itemId++;
					}
				}

				dialogHeader = "Edit demand attributes";
				break;

			case ROUTE:
				long[] routeIds = netPlan.getRouteIdsVector();
				itemIds = routeIds;
				int R = routeIds.length;

				columnNames = new String[R + 1];
				columnNames[0] = "Attribute";
				for (int p = 0; p < R; p++)
				{
					long routeId = routeIds[p];
					columnNames[p + 1] = generateColumnName(type, routeId);
					attributes.addAll(netPlan.getRouteAttributeMap(routeId).keySet());
				}

				if (attributes.isEmpty())
				{
					data = new Object[1][R + 1];
				}
				else
				{
					data = new Object[attributes.size()][R + 1];

					int itemId = 0;
					Iterator<String> it = attributes.iterator();
					while (it.hasNext())
					{
						String attribute = it.next();
						data[itemId][0] = attribute;
						for (int p = 0; p < R; p++)
						{
							long routeId = routeIds[p];
							data[itemId][p + 1] = netPlan.getRouteAttribute(routeId, attribute);
							if (data[itemId][p + 1] == null)
							{
								data[itemId][p + 1] = "";
							}
						}

						itemId++;
					}
				}

				dialogHeader = "Edit route attributes";
				break;

			case PROTECTION_SEGMENT:
				long[] segmentIds = netPlan.getProtectionSegmentIdsVector();
				itemIds = segmentIds;
				int S = segmentIds.length;

				columnNames = new String[S + 1];
				columnNames[0] = "Attribute";
				for (int s = 0; s < S; s++)
				{
					long segmentId = segmentIds[s];
					columnNames[s + 1] = generateColumnName(type, segmentId);
					attributes.addAll(netPlan.getProtectionSegmentAttributeMap(segmentId).keySet());
				}

				if (attributes.isEmpty())
				{
					data = new Object[1][S + 1];
				}
				else
				{
					data = new Object[attributes.size()][S + 1];

					int itemId = 0;
					Iterator<String> it = attributes.iterator();
					while (it.hasNext())
					{
						String attribute = it.next();
						data[itemId][0] = attribute;
						for (int segmentIndex = 0; segmentIndex < S; segmentIndex++)
						{
							long segmentId = segmentIds[segmentIndex];
							data[itemId][segmentIndex + 1] = netPlan.getProtectionSegmentAttribute(segmentId, attribute);
							if (data[itemId][segmentIndex + 1] == null)
							{
								data[itemId][segmentIndex + 1] = "";
							}
						}

						itemId++;
					}
				}

				dialogHeader = "Edit protection segment attributes";
				break;

			case FORWARDING_RULE:
				List<Pair<Long, Long>> forwardingRules = new ArrayList<Pair<Long, Long>>(netPlan.getForwardingRules());
				itemIds = forwardingRules;
				int F = forwardingRules.size();

				columnNames = new String[F + 1];
				columnNames[0] = "Attribute";
				for (int f = 0; f < F; f++)
				{
					Pair<Long, Long> demandLinkPair = forwardingRules.get(f);
					columnNames[f + 1] = generateColumnName(type, demandLinkPair);
					attributes.addAll(netPlan.getForwardingRuleAttributeMap(demandLinkPair).keySet());
				}

				if (attributes.isEmpty())
				{
					data = new Object[1][F + 1];
				}
				else
				{
					data = new Object[attributes.size()][F + 1];

					int itemId = 0;
					Iterator<String> it = attributes.iterator();
					while (it.hasNext())
					{
						String attribute = it.next();
						data[itemId][0] = attribute;
						for (int f = 0; f < F; f++)
						{
							Pair<Long, Long> demandLinkPair = forwardingRules.get(f);
							data[itemId][f + 1] = netPlan.getForwardingRuleAttribute(demandLinkPair, attribute);
							if (data[itemId][f + 1] == null)
							{
								data[itemId][f + 1] = "";
							}
						}

						itemId++;
					}
				}

				dialogHeader = "Edit forwarding rule attributes";
				break;

			case SRG:
				long[] srgIds = netPlan.getSRGIdsVector();
				itemIds = srgIds;
				int numSRGs = srgIds.length;

				columnNames = new String[numSRGs + 1];
				columnNames[0] = "Attribute";
				for (int srgIndex = 0; srgIndex < numSRGs; srgIndex++)
				{
					long srgId = srgIds[srgIndex];
					columnNames[srgIndex + 1] = generateColumnName(type, srgId);
					attributes.addAll(netPlan.getSRGAttributeMap(srgId).keySet());
				}

				if (attributes.isEmpty())
				{
					data = new Object[1][numSRGs + 1];
				}
				else
				{
					data = new Object[attributes.size()][numSRGs + 1];

					int itemId = 0;
					Iterator<String> it = attributes.iterator();
					while (it.hasNext())
					{
						String attribute = it.next();
						data[itemId][0] = attribute;
						for (int srgIndex = 0; srgIndex < numSRGs; srgIndex++)
						{
							long srgId = srgIds[srgIndex];
							data[itemId][srgIndex + 1] = netPlan.getSRGAttribute(srgId, attribute);
							if (data[itemId][srgIndex + 1] == null)
							{
								data[itemId][srgIndex + 1] = "";
							}
						}

						itemId++;
					}
				}

				dialogHeader = "Edit SRG attributes";
				break;

			default:
				throw new IllegalArgumentException("Invalid network element type");
		}

		DefaultTableModel model = new ClassAwareTableModelImpl(data, columnNames, type, netPlan);

		table = new AdvancedJTable(model);
		table.addMouseListener(new PopupAttributeMenu(table, attributes));
		if (attributes.isEmpty()) table.setEnabled(false);

		JScrollPane pane = new JScrollPane(table);
		new FixedColumnDecorator(pane, 1);

		SwingUtils.configureCloseDialogOnEscape(this);
		setTitle(dialogHeader);
		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(pane, BorderLayout.CENTER);
		setModalityType(Dialog.ModalityType.APPLICATION_MODAL);
		pack();
		setLocationRelativeTo(null);
	}
	
	/**
	 * Constructor that allows to show just a single item.
	 *
	 * @param callback Reference to the handler of the network design
	 * @param type Type of element (i.e. layers, nodes, links, and so on)
	 * @param itemId Item identifier
	 * @since 0.3.0
	 */
	public AttributeEditor(final INetworkCallback callback, final NetworkElementType type, Object itemId)
	{
		this(callback, type);
		
		final TableColumnHider tch = new TableColumnHider(table);
		String columnName = generateColumnName(type, itemId);
		tch.maintainOnly(columnName);
		
		final JPanel pane = new JPanel();
		JButton viewAll = new JButton("View all " + type.toString() + "s");
		
		viewAll.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				getContentPane().remove(pane);
				getContentPane().revalidate();
				
				tch.showAll();
			}
		});
		
		pane.add(viewAll);
		getContentPane().add(pane, BorderLayout.NORTH);
	}
			
	@Override
	public void actionPerformed(ActionEvent e)
	{
		dispose();
	}
	
	private String generateColumnName(final NetworkElementType type, Object itemId)
	{
		switch(type)
		{
			case DEMAND:
				return "Demand " + (long) itemId;
				
			case FORWARDING_RULE:
				return "[d" + ((Pair<Long, Long>) itemId).getFirst() + ", e" + ((Pair<Long, Long>) itemId).getSecond() + "]";
				
			case LAYER:
				return "Layer " + (long) itemId;
				
			case LINK:
				return "Link " + (long) itemId;
				
			case NODE:
				return "Node " + (long) itemId;
				
			case PROTECTION_SEGMENT:
				return "Protection segment " + (long) itemId;

			case ROUTE:
				return "Route " + (long) itemId;

			case SRG:
				return "SRG " + (long) itemId;
				
			default:
				throw new RuntimeException("Bad");
		}
	}
	
	private static class PopupAttributeMenu extends MouseAdapter
	{
		private final JTable table;
		private final Set<String> attributes;

		public PopupAttributeMenu(JTable table, Set<String> attributes)
		{
			this.table = table;
			this.attributes = attributes;
		}

		private boolean isTableEmpty()
		{
			return (!(table.getModel().getValueAt(0, 0) instanceof String));
		}

		private void doPopup(MouseEvent e)
		{
			final int row = table.rowAtPoint(e.getPoint());

			JPopupMenu popup = new JPopupMenu();

			JMenuItem addAttribute = new JMenuItem("Add attribute");
			addAttribute.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					String attributeName;

					while (true)
					{
						attributeName = JOptionPane.showInputDialog("Please enter new attribute name: ");
						if (attributeName == null) return;

						try
						{
							if (attributes.contains(attributeName)) throw new RuntimeException("Attribute '" + attributeName + "' already exists");
							if (attributeName.isEmpty()) throw new RuntimeException("Attribute cannot be null or empty");
						}
						catch (Throwable ex)
						{
							ErrorHandling.showWarningDialog(ex.getMessage(), "Error adding attribute");
							continue;
						}

						break;
					}

					int numColumns = table.getModel().getColumnCount();
					String[] newRow = new String[numColumns];
					Arrays.fill(newRow, "");

					newRow[0] = attributeName;
					DefaultTableModel model = (DefaultTableModel) table.getModel();
					model.addRow(newRow);

					if (!table.isEnabled())
					{
						model.removeRow(0);
						table.setEnabled(true);
					}
				}
			});

			popup.add(addAttribute);

			if (!isTableEmpty() && row != -1)
			{
				JMenuItem removeAttribute = new JMenuItem("Remove attribute");
				removeAttribute.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						DefaultTableModel model = (DefaultTableModel) table.getModel();
						int actualRow = table.convertRowIndexToModel(row);
						String attributeName = (String) model.getValueAt(actualRow, 0);

						int rc = JOptionPane.showConfirmDialog(null, "Are you sure?", "Removing attribute '" + attributeName + "'", JOptionPane.YES_NO_OPTION);
						if (rc != JOptionPane.YES_OPTION) return;

						int numColumns = model.getColumnCount();
						for (int columnId = 1; columnId < numColumns; columnId++)
						{
							model.setValueAt("", actualRow, columnId);
						}

						model.removeRow(actualRow);
					}
				});

				JMenuItem setValueToAll = new JMenuItem("Set value attribute for all elements");
				setValueToAll.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						DefaultTableModel model = (DefaultTableModel) table.getModel();
						int actualRow = table.convertRowIndexToModel(row);
						String attributeName = (String) model.getValueAt(actualRow, 0);

						String attributeValue;

						while (true)
						{
							attributeValue = JOptionPane.showInputDialog("Please enter new value for attribute '" + attributeName + "': ");
							if (attributeValue == null) return;

							break;
						}

						int numColumns = model.getColumnCount();
						for (int columnId = 1; columnId < numColumns; columnId++)
						{
							model.setValueAt(attributeValue, actualRow, columnId);
						}
					}
				});

				popup.add(removeAttribute);
				popup.add(setValueToAll);
			}

			popup.show(e.getComponent(), e.getX(), e.getY());
		}

		@Override
		public void mouseClicked(MouseEvent e)
		{
			table.clearSelection();

			if (SwingUtilities.isRightMouseButton(e)) doPopup(e);
		}
	}

	private class ClassAwareTableModelImpl extends ClassAwareTableModel
	{
		private final NetworkElementType type;
		private final NetPlan netPlan;

		public ClassAwareTableModelImpl(Object[][] dataVector, Object[] columnIdentifiers, NetworkElementType type, NetPlan netPlan)
		{
			super(dataVector, columnIdentifiers);
			this.type = type;
			this.netPlan = netPlan;
		}

		@Override
		public boolean isCellEditable(int rowIndex, int columnIndex)
		{
			return columnIndex != 0;
		}

		@Override
		public void setValueAt(Object value, int row, int col)
		{
			String oldValue = getValueAt(row, col).toString();
			String newValue = value.toString();

			if (newValue.equals(oldValue)) return;

			if (col != 0)
			{
				String key = getValueAt(row, 0).toString();

				try
				{
					switch (type)
					{
						case LAYER:
							long layerId = ((long[]) itemIds)[col - 1];
							if (newValue.isEmpty()) netPlan.removeLayerAttribute(layerId, key);
							else netPlan.setLayerAttribute(layerId, key, newValue);
							break;

						case NODE:
							long nodeId = ((long[]) itemIds)[col - 1];
							if (newValue.isEmpty()) netPlan.removeNodeAttribute(nodeId, key);
							else netPlan.setNodeAttribute(nodeId, key, newValue);
							break;

						case LINK:
							long linkId = ((long[]) itemIds)[col - 1];
							if (newValue.isEmpty()) netPlan.removeLinkAttribute(linkId, key);
							else netPlan.setLinkAttribute(linkId, key, newValue);
							break;

						case DEMAND:
							long demandId = ((long[]) itemIds)[col - 1];
							if (newValue.isEmpty()) netPlan.removeDemandAttribute(demandId, key);
							else netPlan.setDemandAttribute(demandId, key, newValue);
							break;

						case ROUTE:
							long routeId = ((long[]) itemIds)[col - 1];
							if (newValue.isEmpty()) netPlan.removeRouteAttribute(routeId, key);
							else netPlan.setRouteAttribute(routeId, key, newValue);
							break;

						case PROTECTION_SEGMENT:
							long segmentId = ((long[]) itemIds)[col - 1];
							if (newValue.isEmpty()) netPlan.removeProtectionSegmentAttribute(segmentId, key);
							else netPlan.setProtectionSegmentAttribute(segmentId, key, newValue);
							break;
							
						case FORWARDING_RULE:
							Pair<Long, Long> demandLinkPair = ((List<Pair<Long, Long>>) itemIds).get(col - 1);
							if (newValue.isEmpty()) netPlan.removeForwardingRuleAttribute(demandLinkPair, key);
							else netPlan.setForwardingRuleAttribute(demandLinkPair, key, newValue);
							break;

						case SRG:
							long srgId = ((long[]) itemIds)[col - 1];
							if (newValue.isEmpty()) netPlan.removeSRGAttribute(srgId, key);
							else netPlan.setSRGAttribute(srgId, key, newValue);
							break;

						default:
							throw new IllegalArgumentException("Invalid network element type");
					}
				}
				catch (Throwable e)
				{
					ErrorHandling.showErrorDialog(e.getMessage(), "Error modifying attributes");
					return;
				}
			}
			
			super.setValueAt(newValue, row, col);
		}
	}
}
