/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.gui.utils;

import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.internal.Constants.DialogType;
import com.net2plan.internal.Constants.IOFeature;
import com.net2plan.internal.SystemUtils;
import com.net2plan.internal.plugins.IOFilter;
import java.awt.Component;
import java.awt.HeadlessException;
import java.io.File;
import java.util.Collection;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;

/**
 * Extends {@link com.net2plan.gui.utils.FileChooserConfirmOverwrite FileChooserConfirmOverwrite} 
 * to support custom IO filters for Net2Plan.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.2
 */
public class FileChooserNetworkDesign extends FileChooserConfirmOverwrite
{
	private boolean finishConfiguration;
	private Set<IOFilter> loadFilters;
	private Set<IOFilter> saveFilters;

	/**
	 * Default constructor.
	 *
	 * @param currentDirectory Current directory
	 * @param type Dialog type
	 * @since 0.3.0
	 */
	public FileChooserNetworkDesign(File currentDirectory, DialogType type)
	{
		super(currentDirectory);
		
		loadFilters = new LinkedHashSet<IOFilter>();
		saveFilters = new LinkedHashSet<IOFilter>();
		
		switch(type)
		{
			case NETWORK_DESIGN:
				loadFilters.addAll(IOFilter.getIOFiltersByFeatures(EnumSet.of(IOFeature.LOAD_DESIGN_SINGLELAYER, IOFeature.LOAD_DESIGN_MULTILAYER)));
				saveFilters.addAll(IOFilter.getIOFiltersByFeatures(EnumSet.of(IOFeature.SAVE_DESIGN_SINGLELAYER, IOFeature.SAVE_DESIGN_MULTILAYER)));
				break;
				
			case DEMANDS:
				loadFilters.addAll(IOFilter.getIOFiltersByFeatures(EnumSet.of(IOFeature.LOAD_DEMANDS)));
				saveFilters.addAll(IOFilter.getIOFiltersByFeatures(EnumSet.of(IOFeature.SAVE_DEMANDS)));
				break;
		}

		finishConfiguration = false;

		super.setAcceptAllFileFilterUsed(false);
		super.setFileSelectionMode(JFileChooser.FILES_ONLY);

		finishConfiguration = true;
	}

	@Override
	public void addChoosableFileFilter(FileFilter filter)
	{
		if (finishConfiguration)
		{
			throw new UnsupportedOperationException("Unsupported operation");
		}
		else
		{
			super.addChoosableFileFilter(filter);
		}
	}

	@Override
	public void setAcceptAllFileFilterUsed(boolean b)
	{
		if (finishConfiguration)
		{
			throw new UnsupportedOperationException("Unsupported operation");
		}
		else
		{
			super.setAcceptAllFileFilterUsed(b);
		}
	}
	
	@Override
	public void setFileSelectionMode(int mode)
	{
		if (finishConfiguration)
		{
			throw new UnsupportedOperationException("Unsupported operation");
		}
		else
		{
			super.setFileSelectionMode(mode);
		}
	}
	
	@Override
    public int showOpenDialog(Component parent) throws HeadlessException
	{
		configureFileChooser(true);
		return super.showOpenDialog(parent);
    }

	@Override
    public int showSaveDialog(Component parent) throws HeadlessException
	{
		configureFileChooser(false);
		return super.showSaveDialog(parent);
    }
	
	private void configureFileChooser(boolean isOpenDialog)
	{
		finishConfiguration = false;
		
		for (FileFilter existingFilter : super.getChoosableFileFilters())
			super.removeChoosableFileFilter(existingFilter);

		boolean isPredefinedFilter = false;

		for (IOFilter filter : isOpenDialog ? loadFilters : saveFilters)
		{
			addChoosableFileFilter(filter);

			if (!isPredefinedFilter)
			{
				super.setFileFilter(filter);
				isPredefinedFilter = true;
			}
		}

		finishConfiguration = true;
	}
	
	private IOFilter getCurrentIOFilter()
	{
		FileFilter currentFilter = getFileFilter();
		if (!(currentFilter instanceof IOFilter))
		{
			throw new RuntimeException("Bad filter");
		}

		return (IOFilter) currentFilter;
	}
	
	/**
	 * Loads the demands from the selected file.
	 *
	 * @return Network design with the demands from the input file
	 * @since 0.3.0
	 */
	public NetPlan readDemands()
	{
		IOFilter importer = getCurrentIOFilter();
		NetPlan netPlan = importer.readDemandSetFromFile(getSelectedFile());
		
		return netPlan;
	}

	/**
	 * Loads a network design from the selected file.
	 *
	 * @return Network design
	 * @since 0.2.2
	 */
	public NetPlan readNetPlan()
	{
		IOFilter importer = getCurrentIOFilter();
		NetPlan netPlan = importer.readFromFile(getSelectedFile());
		return netPlan;
	}

	/**
	 * Loads a network design from the selected file.
	 *
	 * @return Network design
	 * @since 0.2.2
	 */
	public Collection<NetPlan> readNetPlans()
	{
		List<NetPlan> netPlans = new LinkedList<NetPlan>();
		IOFilter importer = getCurrentIOFilter();
		File[] files = getSelectedFiles();
		for (File file : files) netPlans.add(importer.readFromFile(file));
		
		return netPlans;
	}
	
	/**
	 * Saves the demands from the given network design to the selected file.
	 *
	 * @param netPlan Network design
	 * @since 0.3.0
	 */
	public void saveDemands(NetPlan netPlan)
	{
		IOFilter exporter = getCurrentIOFilter();
		File file = getSelectedFile();
		exporter.saveDemandSetToFile(netPlan, file);
	}

	/**
	 * Saves the demands from the given network designs to the selected file.
	 *
	 * @param netPlans Collection of network designs to be saved
	 * @since 0.3.0
	 */
	public void saveDemands(Collection<NetPlan> netPlans)
	{
		if (netPlans.size() == 1)
		{
			saveDemands(netPlans.iterator().next());
			return;
		}
		
		IOFilter exporter = getCurrentIOFilter();
		File file = getSelectedFile();
		
		String path = file.getAbsolutePath();
		String pathSeparator = SystemUtils.getDirectorySeparator();
		String fileName = SystemUtils.getFilename(file);
		String extension = SystemUtils.getExtension(file);
		
		int i = 0;
		Iterator<NetPlan> it = netPlans.iterator();
		while(it.hasNext())
		{
			NetPlan netPlan = it.next();
			
			if (extension.isEmpty())
			{
				exporter.saveDemandSetToFile(netPlan, new File(path + pathSeparator + fileName + "_" + i));
			}
			else
			{
				exporter.saveDemandSetToFile(netPlan, new File(path + pathSeparator + fileName + "_" + i + "." + extension));
			}
			
			i++;
		}
	}

	/**
	 * Saves the given network design to the selected file.
	 *
	 * @param netPlan Network design
	 * @since 0.2.2
	 */
	public void saveNetPlan(NetPlan netPlan)
	{
		IOFilter exporter = getCurrentIOFilter();
		File file = getSelectedFile();
		exporter.saveToFile(netPlan, file);
	}
}
