/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.gui.utils;

import com.net2plan.gui.utils.CellRenderers.UnfocusableCellRenderer;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.internal.ErrorHandling;
import com.net2plan.libraries.SRGUtils;
import com.net2plan.utils.StringUtils;
import com.net2plan.gui.utils.topology.INetworkCallback;
import com.net2plan.gui.utils.topology.TopologyPanel;
import com.net2plan.internal.Constants.NetworkElementType;
import com.net2plan.libraries.IPUtils;
import com.net2plan.libraries.SRGUtils.SharedRiskModel;
import com.net2plan.utils.CollectionUtils;
import com.net2plan.utils.Pair;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;
import javax.swing.WindowConstants;
import javax.swing.border.LineBorder;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;
import net.miginfocom.swing.MigLayout;

/**
 * Class in charge of showing a popup menu in the 'edit network plan' tab of the 
 * Offline Network Design tool.
 * 
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.3
 */
public class PopupMenuNetPlan extends MouseAdapter
{
	private final TableModel model;
	private final NetworkElementType networkElementType;
	private final boolean isEditable;
	private final INetworkCallback callback;
	private final TopologyPanel topologyPanel;

	/**
	 * Default constructor.
	 * 
	 * @param callback
	 * @param topologyPanel
	 * @param model
	 * @param networkElementType
	 * @param isEditable
	 * @since 0.3.0
	 */
	public PopupMenuNetPlan(INetworkCallback callback, TopologyPanel topologyPanel, TableModel model, NetworkElementType networkElementType, boolean isEditable)
	{
		this.callback = callback;
		this.networkElementType = networkElementType;
		this.isEditable = isEditable;
		this.topologyPanel = topologyPanel;
		this.model = model;
	}
	
	@Override
	public void mouseClicked(final MouseEvent e)
	{
		topologyPanel.getCanvas().resetPickedState();
		
		Object auxItemId = null;
		int row = -1;
		if (!isTableEmpty())
		{
			JTable table = getTable(e);
			row = table.rowAtPoint(e.getPoint());
			if (row != -1)
			{
				row = table.convertRowIndexToModel(row);
				if (model.getValueAt(row, 0) == null)
					row = row - 1;
				
				if (networkElementType == NetworkElementType.FORWARDING_RULE)
					auxItemId = Pair.of(Long.parseLong(model.getValueAt(row, 1).toString().split(" ")[0]), Long.parseLong(model.getValueAt(row, 2).toString().split(" ")[0]));
				else
					auxItemId = (Long) model.getValueAt(row, 0);
			}
		}
		
		final Object itemId = auxItemId;
		
		if (SwingUtilities.isRightMouseButton(e))
		{
			doPopup(e, row, itemId);
			return;
		}

		if (itemId == null) return;
		
		SwingUtilities.invokeLater(new Runnable()
		{
			@Override
			public void run()
			{
				showInCanvas(e, itemId);
		    }
		});		
	}

	private static void createForwardingRuleGUI(final INetworkCallback callback)
	{
		final NetPlan netPlan = callback.getDesign();
		final JComboBox nodeSelector = new WiderJComboBox();
		final JComboBox linkSelector = new WiderJComboBox();
		final JComboBox demandSelector = new WiderJComboBox();
		final JTextField txt_splittingRatio = new JTextField(5);
		
		ItemListener nodeListener = new ItemListener()
		{
			@Override
			public void itemStateChanged(ItemEvent e)
			{
				JComboBox me = (JComboBox) e.getSource();
				linkSelector.removeAllItems();
				long nodeId = (long) ((StringLabeller) me.getSelectedItem()).getObject();
				Set<Long> linkIds = netPlan.getNodeOutgoingLinks(nodeId);
				for (long linkId : linkIds)
				{
					final long originNodeId = netPlan.getLinkOriginNode(linkId);
					final String originNodeName = netPlan.getNodeName(originNodeId);
					String originNodeLabel = "Node " + originNodeId;
					if (!originNodeName.isEmpty()) originNodeLabel += " (" + originNodeName + ")";

					final long destinationNodeId = netPlan.getLinkDestinationNode(linkId);
					final String destinationNodeName = netPlan.getNodeName(destinationNodeId);
					String destinationNodeLabel = "Node " + destinationNodeId;
					if (!destinationNodeName.isEmpty()) destinationNodeLabel += " (" + destinationNodeName + ")";

					String linkLabel = "e" + linkId + ": " + originNodeLabel + " -> " + destinationNodeLabel;
					linkSelector.addItem(StringLabeller.of(linkId, linkLabel));
				}
				
				linkSelector.setSelectedIndex(0);
			}
		};
				
		ItemListener linkDemandListener = new ItemListener()
		{
			@Override
			public void itemStateChanged(ItemEvent e)
			{
				long demandId = (long) ((StringLabeller) demandSelector.getSelectedItem()).getObject();
				long linkId = (long) ((StringLabeller) linkSelector.getSelectedItem()).getObject();
				Pair<Long, Long> demandLinkPair = Pair.of(demandId, linkId);
				double splittingRatio;
				if (netPlan.isForwardingRuleActive(demandLinkPair))
				{
					splittingRatio = netPlan.getForwardingRuleSplittingRatio(demandLinkPair);
				}
				else
				{
					long nodeId = netPlan.getLinkOriginNode(linkId);
					Set<Pair<Long, Long>> forwardingRules_thisNode = netPlan.getNodeDemandForwardingRules(nodeId, demandId);
					double totalSplittingRatio = 0;
					for(Pair<Long, Long> demandLinkPair1 : forwardingRules_thisNode)
						totalSplittingRatio += netPlan.getForwardingRuleSplittingRatio(demandLinkPair1);

					splittingRatio = Math.max(0, 1 - totalSplittingRatio);
				}
				
				txt_splittingRatio.setText(Double.toString(splittingRatio));
			}
		};

		nodeSelector.addItemListener(nodeListener);

		Set<Long> nodeIds = netPlan.getNodeIds();
		for (long nodeId : nodeIds)
		{
			if (netPlan.getNodeOutgoingLinks(nodeId).isEmpty()) continue;
			
			final String nodeName = netPlan.getNodeName(nodeId);
			String nodeLabel = "Node " + nodeId;
			if (!nodeName.isEmpty()) nodeLabel += " (" + nodeName + ")";

			nodeSelector.addItem(StringLabeller.of(nodeId, nodeLabel));
		}

		linkSelector.addItemListener(linkDemandListener);		
		demandSelector.addItemListener(linkDemandListener);
		
		Set<Long> demandIds = netPlan.getDemandIds();
		for (long demandId : demandIds)
		{
			final long ingressNodeId = netPlan.getDemandIngressNode(demandId);
			final String ingressNodeName = netPlan.getNodeName(ingressNodeId);
			String ingressNodeLabel = "Node " + ingressNodeId;
			if (!ingressNodeName.isEmpty()) ingressNodeLabel += " (" + ingressNodeName + ")";

			final long egressNodeId = netPlan.getDemandEgressNode(demandId);
			final String egressNodeName = netPlan.getNodeName(egressNodeId);
			String egressNodeLabel = "Node " + egressNodeId;
			if (!egressNodeName.isEmpty()) egressNodeLabel += " (" + egressNodeName + ")";
			
			String demandLabel = "d" + demandId + ": " + ingressNodeLabel + " -> " + egressNodeLabel;
			demandSelector.addItem(StringLabeller.of(demandId, demandLabel));
		}
		
		nodeSelector.setSelectedIndex(0);
		demandSelector.setSelectedIndex(0);
		
		JPanel pane = new JPanel(new MigLayout("fill", "[][grow]", "[][][][][]"));
		pane.add(new JLabel("Node where to install the rule: "));
		pane.add(nodeSelector, "wrap");
		pane.add(new JLabel("Outgoing link: "));
		pane.add(linkSelector, "wrap");
		pane.add(new JLabel("Demand: "));
		pane.add(demandSelector, "wrap");
		pane.add(new JLabel("Splitting ratio: "));
		pane.add(txt_splittingRatio, "wrap");
		
		while (true)
		{
			int result = JOptionPane.showConfirmDialog(null, pane, "Please enter information for the new forwarding rule", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
			if (result != JOptionPane.OK_OPTION) return;

			try
			{
				long demandId = (long) ((StringLabeller) demandSelector.getSelectedItem()).getObject();
				long linkId = (long) ((StringLabeller) linkSelector.getSelectedItem()).getObject();

				double splittingRatio;
				try
				{
					splittingRatio = Double.parseDouble(txt_splittingRatio.getText());
					if (splittingRatio <= 0) throw new RuntimeException();
				}
				catch (Throwable e)
				{
					ErrorHandling.showErrorDialog("Splitting ratio must be a non-negative non-zero number", "Error adding forwarding rule");
					continue;
				}

				netPlan.addForwardingRule(demandId, linkId, splittingRatio, null);
				break;
			}
			catch (Throwable ex)
			{
				ErrorHandling.showErrorDialog(ex.getMessage(), "Error adding forwarding rule");
			}
		}
	}
	
	private static void createLinkDemandGUI(final NetworkElementType networkElementType, final INetworkCallback callback, final TopologyPanel topologyPanel)
	{
		final NetPlan netPlan = callback.getDesign();
		final JComboBox originNodeSelector = new WiderJComboBox();
		final JComboBox destinationNodeSelector = new WiderJComboBox();
		
		Set<Long> nodeIds = netPlan.getNodeIds();
		for (long nodeId : nodeIds)
		{
			final String nodeName = netPlan.getNodeName(nodeId);
			String nodeLabel = "Node " + nodeId;
			if (!nodeName.isEmpty()) nodeLabel += " (" + nodeName + ")";

			originNodeSelector.addItem(StringLabeller.of(nodeId, nodeLabel));
			destinationNodeSelector.addItem(StringLabeller.of(nodeId, nodeLabel));
		}
		
		ItemListener nodeListener = new ItemListener()
		{
			@Override
			public void itemStateChanged(ItemEvent e)
			{
				long originNodeId = (long) ((StringLabeller) originNodeSelector.getSelectedItem()).getObject();
				long destinationNodeId = (long) ((StringLabeller) destinationNodeSelector.getSelectedItem()).getObject();
				Set<Long> nodePair = new LinkedHashSet<Long>();
				nodePair.add(originNodeId);
				nodePair.add(destinationNodeId);
				topologyPanel.getCanvas().showNodes(nodePair);
			}
		};
				
		originNodeSelector.addItemListener(nodeListener);
		destinationNodeSelector.addItemListener(nodeListener);

		originNodeSelector.setSelectedIndex(0);
		destinationNodeSelector.setSelectedIndex(1);

		JPanel pane = new JPanel();
		pane.add(networkElementType == NetworkElementType.LINK ? new JLabel("Origin node: ") : new JLabel("Ingress node: "));
		pane.add(originNodeSelector);
		pane.add(Box.createHorizontalStrut(15));
		pane.add(networkElementType == NetworkElementType.LINK ? new JLabel("Destination node: ") : new JLabel("Egress node: "));
		pane.add(destinationNodeSelector);

		while (true)
		{
			int result = JOptionPane.showConfirmDialog(null, pane, "Please enter end nodes for the new " + networkElementType, JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
			if (result != JOptionPane.OK_OPTION) return;

			try
			{
				long originNodeId = (long) ((StringLabeller) originNodeSelector.getSelectedItem()).getObject();
				long destinationNodeId = (long) ((StringLabeller) destinationNodeSelector.getSelectedItem()).getObject();

				netPlan.checkNodeIsActive(originNodeId);
				netPlan.checkNodeIsActive(destinationNodeId);

				if (networkElementType == NetworkElementType.LINK)
				{
					long linkId = netPlan.addLink(originNodeId, destinationNodeId, 0, netPlan.getNodePairEuclideanDistance(originNodeId, destinationNodeId), null);
					topologyPanel.getCanvas().addLink(linkId, originNodeId, destinationNodeId);
					topologyPanel.getCanvas().refresh();
				}
				else
				{
					netPlan.addDemand(originNodeId, destinationNodeId, 0, null);
				}

				break;
			}
			catch (Throwable ex)
			{
				ErrorHandling.showErrorDialog(ex.getMessage(), "Error adding " + networkElementType);
			}
		}
	}
	
	private static void createProtectionSegmentGUI(final INetworkCallback callback, final TopologyPanel topologyPanel)
	{
		final NetPlan netPlan = callback.getDesign();
		final Set<Long> nodeIds = netPlan.getNodeIds();
		final JComboBox nodeSelector = new WiderJComboBox();

		final List<JComboBox> seqLinks_cmb = new LinkedList<JComboBox>();
		final JPanel seqLinks_pnl = new JPanel();
		seqLinks_pnl.setLayout(new BoxLayout(seqLinks_pnl, BoxLayout.Y_AXIS));

		nodeSelector.addItemListener(new ItemListener()
		{
			@Override
			public void itemStateChanged(ItemEvent e)
			{
				StringLabeller item = (StringLabeller) e.getItem();

				long nodeId = (Long) item.getObject();
				String nodeName = netPlan.getNodeName(nodeId);

				seqLinks_cmb.clear();
				seqLinks_pnl.removeAll();

				Set<Long> outgoingLinks = netPlan.getNodeOutgoingLinks(nodeId);
				final JComboBox firstLink = new WiderJComboBox();
				for (long linkId : outgoingLinks)
				{
					long destinationNodeId = netPlan.getLinkDestinationNode(linkId);
					String destinationNodeName = netPlan.getNodeName(nodeId);
					firstLink.addItem(StringLabeller.of(linkId, String.format("e%d: n%d (%s) => n%d (%s)", linkId, nodeId, nodeName, destinationNodeId, destinationNodeName)));
				}

				firstLink.addItemListener(new ItemListener()
				{
					@Override
					public void itemStateChanged(ItemEvent e)
					{
						JComboBox me = (JComboBox) e.getSource();
						Iterator<JComboBox> it = seqLinks_cmb.iterator();
						while (it.hasNext())
						{
							if (it.next() == me) break;
						}
						while (it.hasNext())
						{
							JComboBox aux = it.next();
							seqLinks_pnl.remove(aux);
							it.remove();
						}

						seqLinks_pnl.revalidate();

						List<Long> seqLinks = new LinkedList<Long>();
						for (JComboBox link : seqLinks_cmb) seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());

						topologyPanel.getCanvas().showRoute(seqLinks);
					}
				});

				setMaxSize(firstLink);

				seqLinks_cmb.add(firstLink);
				seqLinks_pnl.add(firstLink);

				JPanel pane = new JPanel(new FlowLayout());
				JButton addLink_btn = new JButton("Add new link");
				addLink_btn.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						long linkId = (Long) ((StringLabeller) seqLinks_cmb.get(seqLinks_cmb.size() - 1).getSelectedItem()).getObject();
						long destinationNodeId = netPlan.getLinkDestinationNode(linkId);
						String destinationNodeName = netPlan.getNodeName(destinationNodeId);

						Set<Long> outgoingLinks = netPlan.getNodeOutgoingLinks(destinationNodeId);
						if (outgoingLinks.isEmpty())
						{
							ErrorHandling.showErrorDialog("Last node has no outgoing links", "Error");
							return;
						}

						final JComboBox newLink = new WiderJComboBox();
						for (long nextLinkId : outgoingLinks)
						{
							long nextDestinationNodeId = netPlan.getLinkDestinationNode(nextLinkId);
							String nextDestinationNodeName = netPlan.getNodeName(nextDestinationNodeId);
							newLink.addItem(StringLabeller.of(nextLinkId, String.format("e%d: n%d (%s) => n%d (%s)", nextLinkId, destinationNodeId, destinationNodeName, nextDestinationNodeId, nextDestinationNodeName)));
						}

						newLink.addItemListener(new ItemListener()
						{
							@Override
							public void itemStateChanged(ItemEvent e)
							{
								JComboBox me = (JComboBox) e.getSource();
								Iterator<JComboBox> it = seqLinks_cmb.iterator();
								while (it.hasNext()) if (it.next() == me) break;
								
								while (it.hasNext())
								{
									JComboBox aux = it.next();
									seqLinks_pnl.remove(aux);
									it.remove();
								}

								seqLinks_pnl.revalidate();

								List<Long> seqLinks = new LinkedList<Long>();
								for (JComboBox link : seqLinks_cmb)
								{
									seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());
								}

								topologyPanel.getCanvas().showRoute(seqLinks);
							}
						});

						setMaxSize(newLink);

						seqLinks_cmb.add(newLink);
						seqLinks_pnl.add(newLink, seqLinks_pnl.getComponentCount() - 1);
						seqLinks_pnl.revalidate();

						List<Long> seqLinks = new LinkedList<Long>();
						for (JComboBox link : seqLinks_cmb) seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());

						topologyPanel.getCanvas().showRoute(seqLinks);
					}
				});

				pane.add(addLink_btn);

				JButton removeLink_btn = new JButton("Remove last link");
				removeLink_btn.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						if (seqLinks_cmb.size() < 2)
						{
							ErrorHandling.showErrorDialog("Initial link cannot be removed", "Error");
							return;
						}

						JComboBox cmb = seqLinks_cmb.get(seqLinks_cmb.size() - 1);
						seqLinks_cmb.remove(cmb);
						seqLinks_pnl.remove(cmb);
						seqLinks_pnl.revalidate();

						List<Long> seqLinks = new LinkedList<Long>();
						for (JComboBox link : seqLinks_cmb) seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());

						topologyPanel.getCanvas().showRoute(seqLinks);
					}
				});

				pane.add(removeLink_btn);
				seqLinks_pnl.add(pane);

				seqLinks_pnl.revalidate();

				List<Long> seqLinks = new LinkedList<Long>();
				for (JComboBox link : seqLinks_cmb) seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());

				topologyPanel.getCanvas().showRoute(seqLinks);
			}
		});

		for (long nodeId : nodeIds)
		{
			final String nodeName = netPlan.getNodeName(nodeId);
			final Set<Long> outgoingLinks = netPlan.getNodeOutgoingLinks(nodeId);
			if (outgoingLinks.isEmpty()) continue;

			String nodeLabel = "Node " + nodeId;
			if (!nodeName.isEmpty()) nodeLabel += " (" + nodeName + ")";

			nodeSelector.addItem(StringLabeller.of(nodeId, nodeLabel));
		}

		if (nodeSelector.getItemCount() == 0) throw new Net2PlanException("Bad - No node has outgoing links");

		nodeSelector.setSelectedIndex(0);

		final JTextField txt_reservedCapacity = new JTextField();

		final JScrollPane scrollPane = new JScrollPane(seqLinks_pnl, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		scrollPane.setBorder(BorderFactory.createTitledBorder(new LineBorder(Color.BLACK), "Sequence of links"));
		scrollPane.setAlignmentY(JScrollPane.TOP_ALIGNMENT);

		final JPanel pane = new JPanel(new MigLayout("fill", "[][grow]", "[][grow][]"));
		pane.add(new JLabel("Origin node"));
		pane.add(nodeSelector, "grow, wrap, wmin 50");
		pane.add(scrollPane, "grow, spanx 2, wrap");
		pane.add(new JLabel("Reserved capacity"));
		pane.add(txt_reservedCapacity, "grow");
		pane.setPreferredSize(new Dimension(400, 400));

		while (true)
		{
			int result = JOptionPane.showConfirmDialog(null, pane, "Add new protection segment", JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE);
			if (result != JOptionPane.OK_OPTION) break;

			double reservedCapacity;

			try
			{
				reservedCapacity = Double.parseDouble(txt_reservedCapacity.getText());
				if (reservedCapacity < 0) throw new RuntimeException();
			}
			catch (Throwable e)
			{
				ErrorHandling.showErrorDialog("Reserved capacity must be a non-negative number", "Error adding protection segment");
				continue;
			}

			List<Long> seqLinks = new LinkedList<Long>();
			for (JComboBox link : seqLinks_cmb) seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());

			try
			{
				netPlan.addProtectionSegment(seqLinks, reservedCapacity, null);
			}
			catch (Throwable e)
			{
				ErrorHandling.showErrorDialog(e.getMessage(), "Error adding protection segment");
				continue;
			}

			break;
		}

		topologyPanel.getCanvas().resetPickedState();
	}

	private static void createRouteGUI(final INetworkCallback callback, final TopologyPanel topologyPanel)
	{
		final NetPlan netPlan = callback.getDesign();
		final Set<Long> demandIds = netPlan.getDemandIds();
		final JComboBox demandSelector = new WiderJComboBox();

		final JTextField txt_carriedTraffic = new JTextField();
		final JTextField txt_occupiedCapacity = new JTextField();

		final List<JComboBox> seqLinks_cmb = new LinkedList<JComboBox>();
		final JPanel seqLinks_pnl = new JPanel();
		seqLinks_pnl.setLayout(new BoxLayout(seqLinks_pnl, BoxLayout.Y_AXIS));

		demandSelector.addItemListener(new ItemListener()
		{
			@Override
			public void itemStateChanged(ItemEvent e)
			{
				StringLabeller item = (StringLabeller) e.getItem();
				long demandId = (Long) item.getObject();

				double h_d = netPlan.getDemandOfferedTraffic(demandId);
				double r_d = netPlan.getDemandCarriedTraffic(demandId);
				txt_carriedTraffic.setText(Double.toString(Math.max(0, h_d - r_d)));
				txt_occupiedCapacity.setText("");

				seqLinks_cmb.clear();
				seqLinks_pnl.removeAll();

				long ingressNodeId = netPlan.getDemandIngressNode(demandId);
				String ingressNodeName = netPlan.getNodeName(ingressNodeId);

				Set<Long> outgoingLinks = netPlan.getNodeOutgoingLinks(ingressNodeId);
				final JComboBox firstLink = new WiderJComboBox();
				for (long linkId : outgoingLinks)
				{
					long destinationNodeId = netPlan.getLinkDestinationNode(linkId);
					String destinationNodeName = netPlan.getNodeName(destinationNodeId);
					firstLink.addItem(StringLabeller.of(linkId, String.format("e%d: n%d (%s) => n%d (%s)", linkId, ingressNodeId, ingressNodeName, destinationNodeId, destinationNodeName)));
				}

				firstLink.addItemListener(new ItemListener()
				{
					@Override
					public void itemStateChanged(ItemEvent e)
					{
						JComboBox me = (JComboBox) e.getSource();
						Iterator<JComboBox> it = seqLinks_cmb.iterator();
						while (it.hasNext()) if (it.next() == me) break;

						while (it.hasNext())
						{
							JComboBox aux = it.next();
							seqLinks_pnl.remove(aux);
							it.remove();
						}

						seqLinks_pnl.revalidate();

						List<Long> seqLinks = new LinkedList<Long>();
						for (JComboBox link : seqLinks_cmb)
						{
							seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());
						}

						topologyPanel.getCanvas().showRoute(seqLinks);
					}
				});

				setMaxSize(firstLink);

				seqLinks_cmb.add(firstLink);
				seqLinks_pnl.add(firstLink);

				JPanel pane = new JPanel(new FlowLayout());
				JButton addLink_btn = new JButton("Add new link");
				addLink_btn.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						long linkId = (Long) ((StringLabeller) seqLinks_cmb.get(seqLinks_cmb.size() - 1).getSelectedItem()).getObject();
						long destinationNodeId = netPlan.getLinkDestinationNode(linkId);
						String destinationNodeName = netPlan.getNodeName(destinationNodeId);

						Set<Long> outgoingLinks = netPlan.getNodeOutgoingLinks(destinationNodeId);
						if (outgoingLinks.isEmpty())
						{
							ErrorHandling.showErrorDialog("Last node has no outgoing links", "Error");
							return;
						}

						final JComboBox newLink = new WiderJComboBox();
						for (long nextLinkId : outgoingLinks)
						{
							long nextDestinationNodeId = netPlan.getLinkDestinationNode(nextLinkId);
							String nextDestinationNodeName = netPlan.getNodeName(nextDestinationNodeId);
							newLink.addItem(StringLabeller.of(nextLinkId, String.format("e%d: n%d (%s) => n%d (%s)", nextLinkId, destinationNodeId, destinationNodeName, nextDestinationNodeId, nextDestinationNodeName)));
						}

						newLink.addItemListener(new ItemListener()
						{
							@Override
							public void itemStateChanged(ItemEvent e)
							{
								JComboBox me = (JComboBox) e.getSource();
								Iterator<JComboBox> it = seqLinks_cmb.iterator();
								while (it.hasNext()) if (it.next() == me) break;

								while (it.hasNext())
								{
									JComboBox aux = it.next();
									seqLinks_pnl.remove(aux);
									it.remove();
								}

								seqLinks_pnl.revalidate();

								List<Long> seqLinks = new LinkedList<Long>();
								for (JComboBox link : seqLinks_cmb) seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());

								topologyPanel.getCanvas().showRoute(seqLinks);
							}
						});

						setMaxSize(newLink);

						seqLinks_cmb.add(newLink);
						seqLinks_pnl.add(newLink, seqLinks_pnl.getComponentCount() - 1);
						seqLinks_pnl.revalidate();

						List<Long> seqLinks = new LinkedList<Long>();
						for (JComboBox link : seqLinks_cmb)
						{
							seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());
						}

						topologyPanel.getCanvas().showRoute(seqLinks);
					}
				});

				pane.add(addLink_btn);

				JButton removeLink_btn = new JButton("Remove last link");
				removeLink_btn.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						if (seqLinks_cmb.size() < 2)
						{
							ErrorHandling.showErrorDialog("Initial link cannot be removed", "Error");
							return;
						}

						JComboBox cmb = seqLinks_cmb.get(seqLinks_cmb.size() - 1);
						seqLinks_cmb.remove(cmb);
						seqLinks_pnl.remove(cmb);
						seqLinks_pnl.revalidate();

						List<Long> seqLinks = new LinkedList<Long>();
						for (JComboBox link : seqLinks_cmb) seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());

						topologyPanel.getCanvas().showRoute(seqLinks);
					}
				});

				pane.add(removeLink_btn);
				seqLinks_pnl.add(pane);

				seqLinks_pnl.revalidate();

				List<Long> seqLinks = new LinkedList<Long>();
				for (JComboBox link : seqLinks_cmb) seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());

				topologyPanel.getCanvas().showRoute(seqLinks);
			}
		});

		for (long demandId : demandIds)
		{
			long ingressNodeId = netPlan.getDemandIngressNode(demandId);
			long egressNodeId = netPlan.getDemandEgressNode(demandId);

			final String ingressNodeName = netPlan.getNodeName(ingressNodeId);
			final String egressNodeName = netPlan.getNodeName(egressNodeId);

			final Set<Long> outgoingLinks = netPlan.getNodeOutgoingLinks(ingressNodeId);
			if (outgoingLinks.isEmpty()) continue;

			String demandLabel = "Demand " + demandId;
			demandLabel += ": n" + ingressNodeId;
			if (!ingressNodeName.isEmpty()) demandLabel += " (" + ingressNodeName + ")";

			demandLabel += " => n" + egressNodeId;
			if (!egressNodeName.isEmpty()) demandLabel += " (" + egressNodeName + ")";

			double h_d = netPlan.getDemandOfferedTraffic(demandId);
			double r_d = netPlan.getDemandCarriedTraffic(demandId);

			demandLabel += ", offered traffic = " + h_d;
			demandLabel += ", carried traffic = " + r_d;

			demandSelector.addItem(StringLabeller.of(demandId, demandLabel));
		}

		if (demandSelector.getItemCount() == 0) throw new Net2PlanException("Bad - No node has outgoing links");

		demandSelector.setSelectedIndex(0);

		final JScrollPane scrollPane = new JScrollPane(seqLinks_pnl, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		scrollPane.setBorder(BorderFactory.createTitledBorder(new LineBorder(Color.BLACK), "Sequence of links"));
		scrollPane.setAlignmentY(JScrollPane.TOP_ALIGNMENT);

		final JPanel pane = new JPanel(new MigLayout("fill", "[][grow]", "[][grow][]"));
		pane.add(new JLabel("Demand"));
		pane.add(demandSelector, "growx, wrap, wmin 50");
		pane.add(scrollPane, "grow, spanx 2, wrap");
		pane.add(new JLabel("Carried traffic"));
		pane.add(txt_carriedTraffic, "grow, wrap");
		pane.add(new JLabel("Occupied capacity"));
		pane.add(txt_occupiedCapacity, "grow");
		pane.setPreferredSize(new Dimension(400, 400));

		while (true)
		{
			int result = JOptionPane.showConfirmDialog(null, pane, "Add new route", JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE);
			if (result != JOptionPane.OK_OPTION) break;

			long demandId = (Long) ((StringLabeller) demandSelector.getSelectedItem()).getObject();

			double carriedTraffic;
			double occupiedCapacity = -1;

			try
			{
				carriedTraffic = Double.parseDouble(txt_carriedTraffic.getText());
				if (carriedTraffic < 0) throw new RuntimeException();
			}
			catch (Throwable e)
			{
				ErrorHandling.showErrorDialog("Carried traffic must be a non-negative number", "Error adding route");
				continue;
			}

			try
			{
				if (!txt_occupiedCapacity.getText().isEmpty())
				{
					occupiedCapacity = Double.parseDouble(txt_occupiedCapacity.getText());
					if (occupiedCapacity < 0 && occupiedCapacity != -1) throw new RuntimeException();
				}
			}
			catch (Throwable e)
			{
				ErrorHandling.showErrorDialog("Occupied capacity must be a non-negative number, -1 or empty", "Error adding route");
				continue;
			}
			
			List<Long> seqLinks = new LinkedList<Long>();
			for (JComboBox link : seqLinks_cmb) seqLinks.add((Long) ((StringLabeller) link.getSelectedItem()).getObject());

			try
			{
				netPlan.addRoute(demandId, carriedTraffic, occupiedCapacity, seqLinks, null);
			}
			catch (Throwable e)
			{
				ErrorHandling.showErrorDialog(e.getMessage(), "Error adding route");
				continue;
			}

			break;
		}

		topologyPanel.getCanvas().resetPickedState();
	}

	private JTable getTable(MouseEvent e)
	{
		Object src = e.getSource();
		if (src instanceof JTable)
		{
			JTable table = (JTable) src;
			if (table.getModel() != model) throw new RuntimeException("Table model is not valid");
			
			return table;
		}
		
		throw new RuntimeException("Bad - Event source is not a JTable");
	}
	
	private void showInCanvas(MouseEvent e, Object itemId)
	{
		JTable table = getTable(e);
		if (isTableEmpty()) return;

		int clickCount = e.getClickCount();
		switch(clickCount)
		{
			case 1:
				switch(networkElementType)
				{
					case NODE:
						callback.showNode((long) itemId);
						break;
						
					case LINK:
						callback.showLink((long) itemId);
						break;
						
					case DEMAND:
						callback.showDemand((long) itemId);
						break;
						
					case FORWARDING_RULE:
						callback.showForwardingRule((Pair<Long, Long>) itemId);
						break;
						
					case ROUTE:
						callback.showRoute((long) itemId);
						break;
						
					case PROTECTION_SEGMENT:
						callback.showProtectionSegment((long) itemId);
						break;
						
					case SRG:
						callback.showSRG((long) itemId);
						break;
						
					default:
						break;
				}
				
				break;

			case 2:
				int col = table.convertColumnIndexToModel(table.columnAtPoint(e.getPoint()));
				if (col == -1 || col >= table.getColumnCount()) return;

				NetPlan netPlan = callback.getDesign();
				
				switch (networkElementType)
				{
					case LINK:
						switch (col)
						{
							case 2:
								callback.showNode(netPlan.getLinkOriginNode((long) itemId));
								break;
							case 3:
								callback.showNode(netPlan.getLinkDestinationNode((long) itemId));
								break;
							case 5:
								if (netPlan.isLinkCoupled((long) itemId))
								{
									Pair<Long, Long> coupledDemand = netPlan.getLinkCoupledLowerLayerDemand((long) itemId);
									callback.showDemand(coupledDemand.getFirst(), coupledDemand.getSecond());
								}
								break;
								
							default:
								break;
						}
						break;

					case DEMAND:
						Pair<Long, Long> ingressEgressNodePair = netPlan.getDemandIngressEgressNodePair((long) itemId);
						long ingressNodeId = ingressEgressNodePair.getFirst();
						long egressNodeId = ingressEgressNodePair.getSecond();

						switch (col)
						{
							case 1:
								callback.showNode(ingressNodeId);
								break;
							case 2:
								callback.showNode(egressNodeId);
								break;
							case 3:
								if (netPlan.isDemandCoupled((long) itemId))
								{
									Pair<Long, Long> coupledLink = netPlan.getDemandCoupledUpperLayerLink((long) itemId);
									callback.showLink(coupledLink.getFirst(), coupledLink.getSecond());
								}
								break;
							default:
								break;
						}

						break;

					case ROUTE:
						long demandId = netPlan.getRouteDemand((long) itemId);

						switch (col)
						{
							case 1:
								callback.showDemand(demandId);
								break;
							case 2:
								callback.showNode(netPlan.getDemandIngressNode(demandId));
								break;
							case 3:
								callback.showNode(netPlan.getDemandEgressNode(demandId));
								break;
							default:
								break;
						}

						break;

					case FORWARDING_RULE:
						Pair<Long, Long> forwardingRule = (Pair<Long, Long>) itemId;
						switch (col)
						{
							case 0:
								callback.showNode(netPlan.getLinkOriginNode(forwardingRule.getSecond()));
								break;
							case 1:
								callback.showDemand(forwardingRule.getFirst());
								break;
							case 2:
								callback.showLink(forwardingRule.getSecond());
								break;
							default:
								break;
						}

						break;

					default:
						break;
				}
				
				break;
		}
	}

	private boolean isTableEmpty()
	{
		NetPlan netPlan = callback.getDesign();

		switch (networkElementType)
		{
			case LAYER:
				return false;
				
			case NODE:
				return !netPlan.hasNodes();

			case LINK:
				return !netPlan.hasLinks();

			case DEMAND:
				return !netPlan.hasDemands();

			case FORWARDING_RULE:
				return !netPlan.hasForwardingRules();
				
			case ROUTE:
				return !netPlan.hasRoutes();

			case PROTECTION_SEGMENT:
				return !netPlan.hasProtectionSegments();

			case SRG:
				return !netPlan.hasSRGs();

			default:
				throw new RuntimeException("Bad");
		}
	}

	private void doPopup(final MouseEvent e, final int row, final Object itemId)
	{
		final JTable table = getTable(e);

		JPopupMenu popup = new JPopupMenu();

		if (isEditable)
		{
			popup.add(getAddOption());

			List<JMenuItem> extraAddOptions = getExtraAddOptions();
			if (!extraAddOptions.isEmpty())
				for (JMenuItem item : extraAddOptions)
					popup.add(item);
		}

		if (!isTableEmpty())
		{
			if (isEditable)
			{
				if (row != -1)
				{
					if (popup.getSubElements().length > 0) popup.addSeparator();
					
					if (networkElementType == NetworkElementType.LAYER && callback.getDesign().getNumberOfLayers() == 1)
					{
						
					}
					else
					{
						JMenuItem removeItem = new JMenuItem("Remove " + networkElementType);

						removeItem.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								NetPlan netPlan = callback.getDesign();

								try
								{
									switch (networkElementType)
									{
										case LAYER:
											netPlan.removeLayer((long) itemId);
											callback.updateLayerChooser();
											break;
											
										case FORWARDING_RULE:
											netPlan.removeForwardingRule((Pair<Long, Long>) itemId);
											break;

										case NODE:
											netPlan.removeNode((long) itemId);
											topologyPanel.getCanvas().removeNode((long) itemId);
											break;

										case LINK:
											netPlan.removeLink((long) itemId);
											topologyPanel.getCanvas().removeLink((long) itemId);
											break;

										case DEMAND:
											netPlan.removeDemand((long) itemId);
											break;

										case ROUTE:
											netPlan.removeRoute((long) itemId);
											break;

										case PROTECTION_SEGMENT:
											netPlan.removeProtectionSegment((long) itemId);
											break;

										case SRG:
											netPlan.removeSRG((long) itemId);
											break;

										default:
											throw new RuntimeException("Bad");
									}

									callback.updateNetPlanView();
								}
								catch (Throwable ex)
								{
									ErrorHandling.addErrorOrException(ex, PopupMenuNetPlan.class);
									ErrorHandling.showErrorDialog("Unable to remove " + networkElementType);
								}
							}
						});

						popup.add(removeItem);
					}

					JMenuItem addAttribute = new JMenuItem("Add/edit attribute");
					addAttribute.addActionListener(new ActionListener()
					{
						@Override
						public void actionPerformed(ActionEvent e)
						{
							JTextField txt_key = new JTextField(20);
							JTextField txt_value = new JTextField(20);

							JPanel pane = new JPanel();
							pane.add(new JLabel("Attribute: "));
							pane.add(txt_key);
							pane.add(Box.createHorizontalStrut(15));
							pane.add(new JLabel("Value: "));
							pane.add(txt_value);

							NetPlan netPlan = callback.getDesign();

							while (true)
							{
								int result = JOptionPane.showConfirmDialog(null, pane, "Please enter an attribute name and its value", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
								if (result != JOptionPane.OK_OPTION) return;

								try
								{
									if (txt_key.getText().isEmpty()) throw new Exception("Please, insert an attribute name");

									String attribute = txt_key.getText();
									String value = txt_value.getText();

									switch (networkElementType)
									{
										case LAYER:
											netPlan.setLayerAttribute((long) itemId, attribute, value);
											break;
											
										case FORWARDING_RULE:
											netPlan.setForwardingRuleAttribute((Pair<Long, Long>) itemId, attribute, value);
											break;
											
										case NODE:
											netPlan.setNodeAttribute((long) itemId, attribute, value);
											break;

										case LINK:
											netPlan.setLinkAttribute((long) itemId, attribute, value);
											break;

										case DEMAND:
											netPlan.setDemandAttribute((long) itemId, attribute, value);
											break;

										case ROUTE:
											netPlan.setRouteAttribute((long) itemId, attribute, value);
											break;

										case PROTECTION_SEGMENT:
											netPlan.setProtectionSegmentAttribute((long) itemId, attribute, value);
											break;

										case SRG:
											netPlan.setSRGAttribute((long) itemId, attribute, value);
											break;

										default:
											throw new RuntimeException("Bad");
									}

									break;
								}
								catch (Exception ex)
								{
									ErrorHandling.addErrorOrException(ex, PopupMenuNetPlan.class);
									ErrorHandling.showErrorDialog("Error adding/editing attribute");
								}
							}

							try
							{
								callback.updateNetPlanView();
							}
							catch (Exception ex)
							{
								ErrorHandling.addErrorOrException(ex, PopupMenuNetPlan.class);
								ErrorHandling.showErrorDialog("Unable to add attribute to " + networkElementType);
							}
						}
					});

					popup.add(addAttribute);

					JMenuItem viewAttributes = new JMenuItem("View/edit attributes");
					viewAttributes.addActionListener(new ActionListener()
					{
						@Override
						public void actionPerformed(ActionEvent e)
						{
							try
							{
								NetPlan netPlan = callback.getDesign();
								int itemIndex = table.convertRowIndexToModel(row);
								Object itemId;

								switch (networkElementType)
								{
									case FORWARDING_RULE:
										TableModel model = table.getModel();
										itemId = Pair.of((Long) model.getValueAt(itemIndex, 1), (Long) model.getValueAt(itemIndex, 2));
										break;

									case LAYER:
										long[] layerIds = netPlan.getLayerIdsVector();
										itemId = layerIds[itemIndex];
										break;

									case NODE:
										long[] nodeIds = netPlan.getNodeIdsVector();
										itemId = nodeIds[itemIndex];
										break;

									case LINK:
										long[] linkIds = netPlan.getLinkIdsVector();
										itemId = linkIds[itemIndex];
										break;

									case DEMAND:
										long[] demandIds = netPlan.getDemandIdsVector();
										itemId = demandIds[itemIndex];
										break;

									case ROUTE:
										long[] routeIds = netPlan.getRouteIdsVector();
										itemId = routeIds[itemIndex];
										break;

									case PROTECTION_SEGMENT:
										long[] segmentIds = netPlan.getProtectionSegmentIdsVector();
										itemId = segmentIds[itemIndex];
										break;

									case SRG:
										long[] srgIds = netPlan.getSRGIdsVector();
										itemId = srgIds[itemIndex];
										break;
										
									default:
										throw new RuntimeException("Bad");
								}

								JDialog dialog = new AttributeEditor(callback, networkElementType, itemId);
								dialog.setVisible(true);
								callback.updateNetPlanView();
							}
							catch (Throwable ex)
							{
								ErrorHandling.showErrorDialog(ex.getMessage(), "Error modifying attributes");
							}
						}
					});

					popup.add(viewAttributes);

					JMenuItem removeAttribute = new JMenuItem("Remove attribute");

					removeAttribute.addActionListener(new ActionListener()
					{
						@Override
						public void actionPerformed(ActionEvent e)
						{
							NetPlan netPlan = callback.getDesign();

							try
							{
								int itemIndex = table.convertRowIndexToModel(row);
								Object itemId;

								String[] attributeList;

								switch (networkElementType)
								{
									case FORWARDING_RULE:
										TableModel model = table.getModel();
										itemId = Pair.of((Long) model.getValueAt(itemIndex, 1), (Long) model.getValueAt(itemIndex, 2));
										attributeList = StringUtils.toArray(netPlan.getForwardingRuleAttributeMap((Pair<Long, Long>) itemId).keySet());
										break;

									case LAYER:
										long[] layerIds = netPlan.getLayerIdsVector();
										itemId = layerIds[itemIndex];
										attributeList = StringUtils.toArray(netPlan.getLayerAttributeMap((long) itemId).keySet());
										break;

									case NODE:
										long[] nodeIds = netPlan.getNodeIdsVector();
										itemId = nodeIds[itemIndex];
										attributeList = StringUtils.toArray(netPlan.getNodeAttributeMap((long) itemId).keySet());
										break;

									case LINK:
										long[] linkIds = netPlan.getLinkIdsVector();
										itemId = linkIds[itemIndex];
										attributeList = StringUtils.toArray(netPlan.getLinkAttributeMap((long) itemId).keySet());
										break;

									case DEMAND:
										long[] demandIds = netPlan.getDemandIdsVector();
										itemId = demandIds[itemIndex];
										attributeList = StringUtils.toArray(netPlan.getDemandAttributeMap((long) itemId).keySet());
										break;

									case ROUTE:
										long[] routeIds = netPlan.getRouteIdsVector();
										itemId = routeIds[itemIndex];
										attributeList = StringUtils.toArray(netPlan.getRouteAttributeMap((long) itemId).keySet());
										break;

									case PROTECTION_SEGMENT:
										long[] segmentIds = netPlan.getProtectionSegmentIdsVector();
										itemId = segmentIds[itemIndex];
										attributeList = StringUtils.toArray(netPlan.getProtectionSegmentAttributeMap((long) itemId).keySet());
										break;

									case SRG:
										long[] srgIds = netPlan.getSRGIdsVector();
										itemId = srgIds[itemIndex];
										attributeList = StringUtils.toArray(netPlan.getSRGAttributeMap((long) itemId).keySet());
										break;

									default:
										throw new RuntimeException("Bad");
								}

								if (attributeList.length == 0) throw new Exception("No attribute to remove");

								Object out = JOptionPane.showInputDialog(null, "Please, select an attribute to remove", "Remove attribute", JOptionPane.QUESTION_MESSAGE, null, attributeList, attributeList[0]);
								if (out == null) return;

								String attributeToRemove = out.toString();
								switch (networkElementType)
								{
									case FORWARDING_RULE:
										netPlan.removeForwardingRuleAttribute((Pair<Long, Long>) itemId, attributeToRemove);
										break;

									case LAYER:
										netPlan.removeLayerAttribute((long) itemId, attributeToRemove);
										break;

									case NODE:
										netPlan.removeNodeAttribute((long) itemId, attributeToRemove);
										break;

									case LINK:
										netPlan.removeLinkAttribute((long) itemId, attributeToRemove);
										break;

									case DEMAND:
										netPlan.removeDemandAttribute((long) itemId, attributeToRemove);
										break;

									case ROUTE:
										netPlan.removeRouteAttribute((long) itemId, attributeToRemove);
										break;

									case PROTECTION_SEGMENT:
										netPlan.removeProtectionSegmentAttribute((long) itemId, attributeToRemove);
										break;

									case SRG:
										netPlan.removeSRGAttribute((long) itemId, attributeToRemove);
										break;

									default:
										throw new RuntimeException("Bad");
								}

								callback.updateNetPlanView();
							}
							catch (Throwable ex)
							{
								ErrorHandling.showErrorDialog(ex.getMessage(), "Error removing attribute");
							}
						}
					});

					popup.add(removeAttribute);
				}

				if (popup.getSubElements().length > 0) popup.addSeparator();

				JMenuItem addAttributeAll = new JMenuItem("Add/edit attribute to all");
				addAttributeAll.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						JTextField txt_key = new JTextField(20);
						JTextField txt_value = new JTextField(20);

						JPanel pane = new JPanel();
						pane.add(new JLabel("Attribute: "));
						pane.add(txt_key);
						pane.add(Box.createHorizontalStrut(15));
						pane.add(new JLabel("Value: "));
						pane.add(txt_value);

						NetPlan netPlan = callback.getDesign();

						while (true)
						{
							int result = JOptionPane.showConfirmDialog(null, pane, "Please enter an attribute name and its value", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
							if (result != JOptionPane.OK_OPTION) return;

							try
							{
								if (txt_key.getText().isEmpty()) throw new Exception("Please, insert an attribute name");

								String attribute = txt_key.getText();
								String value = txt_value.getText();

								switch (networkElementType)
								{
									case FORWARDING_RULE:
										Set<Pair<Long, Long>> forwardingRules = netPlan.getForwardingRules();
										for(Pair<Long, Long> forwardingRule : forwardingRules) netPlan.setForwardingRuleAttribute(forwardingRule, attribute, value);

										break;

									case LAYER:
										Set<Long> layerIds = netPlan.getLayerIds();
										for (long layerId : layerIds) netPlan.setLayerAttribute(layerId, attribute, value);

										break;

									case NODE:
										Set<Long> nodeIds = netPlan.getNodeIds();
										for (long nodeId : nodeIds) netPlan.setNodeAttribute(nodeId, attribute, value);

										break;

									case LINK:
										Set<Long> linkIds = netPlan.getLinkIds();
										for (long linkId : linkIds) netPlan.setLinkAttribute(linkId, attribute, value);

										break;

									case DEMAND:
										Set<Long> demandIds = netPlan.getDemandIds();
										for (long demandId : demandIds) netPlan.setDemandAttribute(demandId, attribute, value);

										break;

									case ROUTE:
										Set<Long> routeIds = netPlan.getRouteIds();
										for (long routeId : routeIds) netPlan.setRouteAttribute(routeId, attribute, value);

										break;

									case PROTECTION_SEGMENT:
										Set<Long> segmentIds = netPlan.getProtectionSegmentIds();
										for (long segmentId : segmentIds) netPlan.setProtectionSegmentAttribute(segmentId, attribute, value);

										break;

									case SRG:
										Set<Long> srgIds = netPlan.getSRGIds();
										for (long srgId : srgIds) netPlan.setSRGAttribute(srgId, attribute, value);

										break;

									default:
										throw new RuntimeException("Bad");
								}

								break;
							}
							catch (Exception ex)
							{
								ErrorHandling.showErrorDialog(ex.getMessage(), "Error adding/editing attribute to all " + networkElementType + "s");
							}
						}

						try
						{
							callback.updateNetPlanView();
						}
						catch (Exception ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to add attribute to all nodes");
						}
					}
				});

				popup.add(addAttributeAll);
				
				JMenuItem viewAttributes = new JMenuItem("View/edit attributes from all");
				viewAttributes.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						try
						{
							JDialog dialog = new AttributeEditor(callback, networkElementType);
							dialog.setVisible(true);
							callback.updateNetPlanView();
						}
						catch (Throwable ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Error modifying attributes");
						}
					}
				});

				popup.add(viewAttributes);

				JMenuItem removeAttributeAll = new JMenuItem("Remove attribute from all " + networkElementType + "s");

				removeAttributeAll.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						NetPlan netPlan = callback.getDesign();

						try
						{
							Set<String> attributeSet = new LinkedHashSet<String>();
							Object itemIds;

							switch (networkElementType)
							{
								case FORWARDING_RULE:
									itemIds = netPlan.getForwardingRules();
									for(Pair<Long, Long> forwardingRule : (Set<Pair<Long, Long>>) itemIds) attributeSet.addAll(netPlan.getForwardingRuleAttributeMap(forwardingRule).keySet());

									break;
									
								case LAYER:
									itemIds = netPlan.getLayerIds();
									for (long layerId : (Set<Long>) itemIds) attributeSet.addAll(netPlan.getLayerAttributeMap(layerId).keySet());

									break;

								case NODE:
									itemIds = netPlan.getNodeIds();
									for (long nodeId : (Set<Long>) itemIds) attributeSet.addAll(netPlan.getNodeAttributeMap(nodeId).keySet());

									break;

								case LINK:
									itemIds = netPlan.getLinkIds();
									for (long linkId : (Set<Long>) itemIds) attributeSet.addAll(netPlan.getLinkAttributeMap(linkId).keySet());

									break;

								case DEMAND:
									itemIds = netPlan.getDemandIds();
									for (long demandId : (Set<Long>) itemIds) attributeSet.addAll(netPlan.getDemandAttributeMap(demandId).keySet());

									break;

								case ROUTE:
									itemIds = netPlan.getRouteIds();
									for (long routeId : (Set<Long>) itemIds) attributeSet.addAll(netPlan.getRouteAttributeMap(routeId).keySet());

									break;

								case PROTECTION_SEGMENT:
									itemIds = netPlan.getProtectionSegmentIds();
									for (long segmentId : (Set<Long>) itemIds) attributeSet.addAll(netPlan.getProtectionSegmentAttributeMap(segmentId).keySet());

									break;

								case SRG:
									itemIds = netPlan.getSRGIds();
									for (long srgId : (Set<Long>) itemIds) attributeSet.addAll(netPlan.getSRGAttributeMap(srgId).keySet());

									break;

								default:
									throw new RuntimeException("Bad");
							}

							if (attributeSet.isEmpty()) throw new Exception("No attribute to remove");

							Object out = JOptionPane.showInputDialog(null, "Please, select an attribute to remove", "Remove attribute from all nodes", JOptionPane.QUESTION_MESSAGE, null, attributeSet.toArray(new String[attributeSet.size()]), attributeSet.iterator().next());
							if (out == null) return;

							String attributeToRemove = out.toString();

							switch (networkElementType)
							{
								case FORWARDING_RULE:
									for(Pair<Long, Long> forwardingRule : (Set<Pair<Long, Long>>) itemIds) netPlan.removeForwardingRuleAttribute(forwardingRule, attributeToRemove);
									break;
									
								case LAYER:
									for (long layerId : (Set<Long>) itemIds) netPlan.removeLayerAttribute(layerId, attributeToRemove);
									break;

								case NODE:
									for (long nodeId : (Set<Long>) itemIds) netPlan.removeNodeAttribute(nodeId, attributeToRemove);
									break;

								case LINK:
									for (long linkId : (Set<Long>) itemIds) netPlan.removeLinkAttribute(linkId, attributeToRemove);
									break;

								case DEMAND:
									for (long demandId : (Set<Long>) itemIds) netPlan.removeDemandAttribute(demandId, attributeToRemove);
									break;

								case ROUTE:
									for (long routeId : (Set<Long>) itemIds) netPlan.removeRouteAttribute(routeId, attributeToRemove);
									break;

								case PROTECTION_SEGMENT:
									for (long segmentId : (Set<Long>) itemIds) netPlan.removeProtectionSegmentAttribute(segmentId, attributeToRemove);
									break;

								case SRG:
									for (long srgId : (Set<Long>) itemIds) netPlan.removeSRGAttribute(srgId, attributeToRemove);
									break;

								default:
									throw new RuntimeException("Bad");
							}

							callback.updateNetPlanView();
						}
						catch (Throwable ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Error removing attribute from all " + networkElementType + "s");
						}
					}
				});

				popup.add(removeAttributeAll);

				JMenuItem removeAttributes = new JMenuItem("Remove all attributes from all " + networkElementType + "s");

				removeAttributes.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						NetPlan netPlan = callback.getDesign();

						try
						{
							switch (networkElementType)
							{
								case FORWARDING_RULE:
									Set<Pair<Long, Long>> forwardingRules = netPlan.getForwardingRules();
									for(Pair<Long, Long> forwardingRule : forwardingRules) netPlan.setForwardingRuleAttributeMap(forwardingRule, null);

									break;

								case LAYER:
									Set<Long> layerIds = netPlan.getLayerIds();
									for (long layerId : layerIds) netPlan.setLayerAttributeMap(layerId, null);

									break;

								case NODE:
									Set<Long> nodeIds = netPlan.getNodeIds();
									for (long nodeId : nodeIds) netPlan.setNodeAttributeMap(nodeId, null);

									break;

								case LINK:
									Set<Long> linkIds = netPlan.getLinkIds();
									for (long linkId : linkIds) netPlan.setLinkAttributeMap(linkId, null);

									break;

								case DEMAND:
									Set<Long> demandIds = netPlan.getDemandIds();
									for (long demandId : demandIds) netPlan.setDemandAttributeMap(demandId, null);

									break;

								case ROUTE:
									Set<Long> routeIds = netPlan.getRouteIds();
									for (long routeId : routeIds) netPlan.setRouteAttributeMap(routeId, null);

									break;

								case PROTECTION_SEGMENT:
									Set<Long> segmentIds = netPlan.getProtectionSegmentIds();
									for (long segmentId : segmentIds) netPlan.setProtectionSegmentAttributeMap(segmentId, null);

									break;

								case SRG:
									Set<Long> srgIds = netPlan.getSRGIds();
									for (long srgId : srgIds) netPlan.setSRGAttributeMap(srgId, null);

									break;

								default:
									throw new RuntimeException("Bad");
							}

							callback.updateNetPlanView();
						}
						catch (Throwable ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Error removing attributes");
						}
					}
				});

				popup.add(removeAttributes);

				if (networkElementType != NetworkElementType.LAYER)
				{
					JMenuItem removeItems = new JMenuItem("Remove all " + networkElementType + "s");

					removeItems.addActionListener(new ActionListener()
					{
						@Override
						public void actionPerformed(ActionEvent e)
						{
							NetPlan netPlan = callback.getDesign();

							try
							{
								switch (networkElementType)
								{
									case NODE:
										netPlan.removeAllNodes();
										topologyPanel.getCanvas().updateTopology(netPlan);

										break;

									case LINK:
										long[] linkIds = netPlan.getLinkIdsVector();
										netPlan.removeAllLinks();
										
										for(long linkId : linkIds)
											topologyPanel.getCanvas().removeLink(linkId);
										
										topologyPanel.getCanvas().refresh();

										break;

									case DEMAND:
										netPlan.removeAllDemands();
										break;
										
									case FORWARDING_RULE:
										netPlan.removeAllForwardingRules();
										break;

									case ROUTE:
										netPlan.removeAllRoutes();
										break;

									case PROTECTION_SEGMENT:
										netPlan.removeAllProtectionSegments();
										break;

									case SRG:
										netPlan.removeAllSRGs();
										break;

									default:
										throw new RuntimeException("Bad");
								}

								callback.updateNetPlanView();
							}
							catch (Throwable ex)
							{
								ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to remove all " + networkElementType + "s");
							}
						}
					});

					popup.add(removeItems);
				}

				List<JMenuItem> extraOptions = getExtraOptions(row, itemId);
				if (!extraOptions.isEmpty())
				{
					popup.addSeparator();
					for (JMenuItem item : extraOptions) popup.add(item);
				}
			}
			
			List<JMenuItem> forcedOptions = getForcedOptions();
			if (!forcedOptions.isEmpty())
			{
				if (popup.getSubElements().length > 0) popup.addSeparator();
				for (JMenuItem item : forcedOptions) popup.add(item);
			}
		}

		popup.show(e.getComponent(), e.getX(), e.getY());
	}

	private JMenuItem getAddOption()
	{
		JMenuItem addItem;

		switch (networkElementType)
		{
			case FORWARDING_RULE:
			case LAYER:
			case NODE:
			case LINK:
			case DEMAND:
			case ROUTE:
			case PROTECTION_SEGMENT:
			case SRG:
				addItem = new JMenuItem("Add " + networkElementType);
				break;

			default:
				throw new RuntimeException("Bad");
		}

		addItem.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				NetPlan netPlan = callback.getDesign();

				try
				{
					switch (networkElementType)
					{
						case LAYER:
							netPlan.addLayer(null, null, null, null, null);
							callback.updateLayerChooser();
							break;
							
						case SRG:
							netPlan.addSRG(8748, 12, null);
							callback.updateNetPlanView();
							break;

						case NODE:
							long nodeId = netPlan.addNode(0, 0, null, null);
							String nodeName = netPlan.getNodeName(nodeId);
							Point2D nodeXYPosition = netPlan.getNodeXYPosition(nodeId);
							topologyPanel.getCanvas().addNode(nodeId, nodeXYPosition, nodeName);
							topologyPanel.getCanvas().refresh();
							callback.updateNetPlanView();
							break;

						case FORWARDING_RULE:
							createForwardingRuleGUI(callback);
							callback.updateNetPlanView();
							break;

						case LINK:
						case DEMAND:
							createLinkDemandGUI(networkElementType, callback, topologyPanel);
							callback.updateNetPlanView();
							break;

						case ROUTE:
							createRouteGUI(callback, topologyPanel);
							callback.updateNetPlanView();
							break;

						case PROTECTION_SEGMENT:
							createProtectionSegmentGUI(callback, topologyPanel);
							callback.updateNetPlanView();
							break;

						default:
							throw new RuntimeException("Bad");
					}
				}
				catch (Throwable ex)
				{
					ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to add " + networkElementType);
				}
			}
		});
		
		switch (networkElementType)
		{
			case LINK:
			case DEMAND:
			case PROTECTION_SEGMENT:
			{
				NetPlan netPlan = callback.getDesign();
				int N = netPlan.getNumberOfNodes();
				if (N < 2) addItem.setEnabled(false);

				break;
			}
			
			case ROUTE:
			{
				NetPlan netPlan = callback.getDesign();
				if (!netPlan.hasDemands()) addItem.setEnabled(false);

				break;
			}
			
			case FORWARDING_RULE:
			{
				NetPlan netPlan = callback.getDesign();
				if (!netPlan.hasLinks() || !netPlan.hasDemands()) addItem.setEnabled(false);
				
				break;
			}
			
			default:
				break;
		}

		return addItem;
	}

	private List<JMenuItem> getExtraOptions(final int row, final Object itemId)
	{
		List<JMenuItem> options = new LinkedList<JMenuItem>();
		final NetPlan netPlan = callback.getDesign();

		switch (networkElementType)
		{
			case LAYER:
			case NODE:
			case PROTECTION_SEGMENT:
			case FORWARDING_RULE:
				break;

			case DEMAND:
				if (itemId != null && netPlan.isMultiLayer())
				{
					final long demandId = (long) itemId;
					if (netPlan.isDemandCoupled(demandId))
					{
						JMenuItem decoupleDemandItem = new JMenuItem("Decouple demand");
						decoupleDemandItem.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								netPlan.decoupleDemand(demandId);
								model.setValueAt("", row, 3);
								callback.updateWarnings();
							}
						});

						options.add(decoupleDemandItem);
					}
					else
					{
						JMenuItem createUpperLayerLinkFromDemandItem = new JMenuItem("Create upper layer link from demand");
						createUpperLayerLinkFromDemandItem.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								Set<Long> layerIds = netPlan.getLayerIds();
								final JComboBox layerSelector = new WiderJComboBox();
								for (long layerId : layerIds)
								{
									if (layerId == netPlan.getLayerDefaultId()) continue;
									
									final String layerName = netPlan.getLayerName(layerId);
									String layerLabel = "Layer " + layerId;
									if (!layerName.isEmpty()) layerLabel += " (" + layerName + ")";

									layerSelector.addItem(StringLabeller.of(layerId, layerLabel));
								}

								layerSelector.setSelectedIndex(0);

								JPanel pane = new JPanel();
								pane.add(new JLabel("Select layer: "));
								pane.add(layerSelector);

								while (true)
								{
									int result = JOptionPane.showConfirmDialog(null, pane, "Please select the upper layer to create the link", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
									if (result != JOptionPane.OK_OPTION) return;

									try
									{
										long layerId = (long) ((StringLabeller) layerSelector.getSelectedItem()).getObject();
										netPlan.createUpperLayerLinkFromDemand(demandId, layerId);

										callback.updateNetPlanView();
										break;
									}
									catch (Throwable ex)
									{
										ErrorHandling.showErrorDialog(ex.getMessage(), "Error creating upper layer link from demand");
									}
								}
							}
						});

						options.add(createUpperLayerLinkFromDemandItem);

						JMenuItem coupleDemandToLink = new JMenuItem("Couple demand to upper layer link");
						coupleDemandToLink.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								Set<Long> layerIds = netPlan.getLayerIds();
								final JComboBox layerSelector = new WiderJComboBox();
								final JComboBox linkSelector = new WiderJComboBox();
								for (long layerId : layerIds)
								{
									if (layerId == netPlan.getLayerDefaultId()) continue;
									
									final String layerName = netPlan.getLayerName(layerId);
									String layerLabel = "Layer " + layerId;
									if (!layerName.isEmpty()) layerLabel += " (" + layerName + ")";

									layerSelector.addItem(StringLabeller.of(layerId, layerLabel));
								}
								
								layerSelector.addItemListener(new ItemListener()
								{
									@Override
									public void itemStateChanged(ItemEvent e)
									{
										if (layerSelector.getSelectedIndex() >= 0)
										{
											long selectedLayerId = (Long) ((StringLabeller) layerSelector.getSelectedItem()).getObject();

											linkSelector.removeAllItems();
											Set<Long> linkIds_thisLayer = netPlan.getLinkIds(selectedLayerId);
											for(long linkId : linkIds_thisLayer)
											{
												if (netPlan.isLinkCoupled(selectedLayerId, linkId)) continue;

												long originNodeId = netPlan.getLinkOriginNode(selectedLayerId, linkId);
												long destinationNodeId = netPlan.getLinkDestinationNode(selectedLayerId, linkId);
												String originNodeName = netPlan.getNodeName(originNodeId);
												String destinationNodeName = netPlan.getNodeName(destinationNodeId);

												linkSelector.addItem(new StringLabeller(linkId, "e" + linkId + " [n" + originNodeId + " (" + originNodeName + ") -> n" + destinationNodeId + " (" + destinationNodeName + ")]"));
											}
										}
										
										if (linkSelector.getItemCount() == 0)
										{
											linkSelector.setEnabled(false);
										}
										else 
										{
											linkSelector.setSelectedIndex(0);
											linkSelector.setEnabled(true);
										}
									}
								});

								layerSelector.setSelectedIndex(-1);
								layerSelector.setSelectedIndex(0);

								JPanel pane = new JPanel(new MigLayout("", "[][grow]", "[][]"));
								pane.add(new JLabel("Select layer: "));
								pane.add(layerSelector, "growx, wrap");
								pane.add(new JLabel("Select link: "));
								pane.add(linkSelector, "growx, wrap");

								while (true)
								{
									int result = JOptionPane.showConfirmDialog(null, pane, "Please select the upper layer link", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
									if (result != JOptionPane.OK_OPTION) return;

									try
									{
										long layerId = (long) ((StringLabeller) layerSelector.getSelectedItem()).getObject();
										long linkId;
										try { linkId = (long) ((StringLabeller) linkSelector.getSelectedItem()).getObject(); }
										catch(Throwable ex) { throw new RuntimeException("No link was selected"); }
										
										netPlan.coupleDemandWithUpperLayerLink(demandId, layerId, linkId);

										callback.updateNetPlanView();
										break;
									}
									catch (Throwable ex)
									{
										ErrorHandling.showErrorDialog(ex.getMessage(), "Error coupling upper layer link to demand");
									}
								}
							}
						});

						options.add(coupleDemandToLink);
					}

					final Set<Long> coupledDemands = netPlan.getDemandsCoupledLinkMap().keySet();
					if (!coupledDemands.isEmpty())
					{
						JMenuItem decoupleAllDemandsItem = new JMenuItem("Decouple all demands");
						decoupleAllDemandsItem.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								for(long demandId : new LinkedHashSet<Long>(coupledDemands)) netPlan.decoupleDemand(demandId);

								int numRows = model.getRowCount();
								for(int i = 0; i < numRows; i++) model.setValueAt("", i, 3);

								callback.updateWarnings();
							}
						});

						options.add(decoupleAllDemandsItem);
					}
					
					if (coupledDemands.size() < netPlan.getNumberOfDemands())
					{
						JMenuItem createUpperLayerLinksFromDemandsItem = new JMenuItem("Create upper layer links from uncoupled demands");
						createUpperLayerLinksFromDemandsItem.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								Set<Long> layerIds = netPlan.getLayerIds();
								final JComboBox layerSelector = new WiderJComboBox();
								for (long layerId : layerIds)
								{
									if (layerId == netPlan.getLayerDefaultId()) continue;
									
									final String layerName = netPlan.getLayerName(layerId);
									String layerLabel = "Layer " + layerId;
									if (!layerName.isEmpty()) layerLabel += " (" + layerName + ")";

									layerSelector.addItem(StringLabeller.of(layerId, layerLabel));
								}

								layerSelector.setSelectedIndex(0);

								JPanel pane = new JPanel();
								pane.add(new JLabel("Select layer: "));
								pane.add(layerSelector);

								while (true)
								{
									int result = JOptionPane.showConfirmDialog(null, pane, "Please select the upper layer to create links", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
									if (result != JOptionPane.OK_OPTION) return;

									try
									{
										long layerId = (long) ((StringLabeller) layerSelector.getSelectedItem()).getObject();
										netPlan.createUpperLayerLinksFromDemands(layerId);

										callback.updateNetPlanView();
										break;
									}
									catch (Throwable ex)
									{
										ErrorHandling.showErrorDialog(ex.getMessage(), "Error creating upper layer links");
									}
								}
							}
						});

						options.add(createUpperLayerLinksFromDemandsItem);
					}
				}
				
				break;
				
			case ROUTE:

				if (itemId != null)
				{
					JMenuItem viewEditProtectionSegments = new JMenuItem("View/edit backup segment list");
					viewEditProtectionSegments.addActionListener(new ActionListener()
					{
						@Override
						public void actionPerformed(ActionEvent e)
						{
							try { viewEditBackupSegmentListGUI(callback, topologyPanel, (long) itemId); }
							catch (Exception ex) { ErrorHandling.showErrorDialog(ex.getMessage(), "Error viewing/editing backup segment list"); }
						}
					});

					options.add(viewEditProtectionSegments);
				}

				break;

			case SRG:
				JMenuItem mttfValue = new JMenuItem("Set MTTF to all");

				mttfValue.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						double mttf;

						while (true)
						{
							String str = JOptionPane.showInputDialog(null, "MTTF (in hours, zero or negative value means no failure)", "Set MTTF to all SRGs", JOptionPane.QUESTION_MESSAGE);
							if (str == null) return;

							try
							{
								mttf = Double.parseDouble(str);
								break;
							}
							catch (NumberFormatException ex)
							{
								ErrorHandling.showErrorDialog("Non-valid MTTF value", "Error setting MTTF value");
							}
							catch (Exception ex)
							{
								ErrorHandling.showErrorDialog(ex.getMessage(), "Error setting MTTF");
							}
						}

						NetPlan netPlan = callback.getDesign();

						try
						{
							Set<Long> srgIds = netPlan.getSRGIds();
							for (long srgId : srgIds) netPlan.setSRGMeanTimeToFailInHours(srgId, mttf);

							callback.updateNetPlanView();
						}
						catch (Exception ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to set MTTF to all SRGs");
						}
					}
				});

				options.add(mttfValue);

				JMenuItem mttrValue = new JMenuItem("Set MTTR to all");

				mttrValue.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						double mttr;

						while (true)
						{
							String str = JOptionPane.showInputDialog(null, "MTTR (in hours)", "Set MTTR to all SRGs", JOptionPane.QUESTION_MESSAGE);
							if (str == null) return;

							try
							{
								mttr = Double.parseDouble(str);
								if (mttr <= 0) throw new NumberFormatException();

								break;
							}
							catch (NumberFormatException ex)
							{
								ErrorHandling.showErrorDialog("Non-valid MTTR value. Please, introduce a non-zero non-negative number", "Error setting MTTR value");
							}
							catch (Exception ex)
							{
								ErrorHandling.showErrorDialog(ex.getMessage(), "Error setting MTTR");
							}
						}

						NetPlan netPlan = callback.getDesign();

						try
						{
							Set<Long> srgIds = netPlan.getSRGIds();
							for (long srgId : srgIds) netPlan.setSRGMeanTimeToRepairInHours(srgId, mttr);

							callback.updateNetPlanView();
						}
						catch (Exception ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to set MTTR to all SRGs");
						}
					}
				});

				options.add(mttrValue);

				break;

			case LINK:

				JMenuItem caFixValue = new JMenuItem("Set capacity value to all");

				caFixValue.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						double u_e;

						while (true)
						{
							String str = JOptionPane.showInputDialog(null, "Capacity value", "Set capacity value to all links", JOptionPane.QUESTION_MESSAGE);
							if (str == null) return;

							try
							{
								u_e = Double.parseDouble(str);
								if (u_e < 0) throw new NumberFormatException();

								break;
							}
							catch (NumberFormatException ex)
							{
								ErrorHandling.showErrorDialog("Non-valid capacity value. Please, introduce a non-negative number", "Error setting capacity value");
							}
						}

						NetPlan netPlan = callback.getDesign();
						
						try
						{
							Set<Long> linkIds = netPlan.getLinkIds();
							for(long linkId : linkIds) netPlan.setLinkCapacity(linkId, u_e);

							callback.updateNetPlanView();
						}
						catch (Exception ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to set capacity value to all links");
						}
					}
				});

				options.add(caFixValue);

				JMenuItem lengthToAll = new JMenuItem("Set link length value to all");

				lengthToAll.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						double l_e;

						while (true)
						{
							String str = JOptionPane.showInputDialog(null, "Link length value (in km)", "Set link length value to all links", JOptionPane.QUESTION_MESSAGE);
							if (str == null) return;

							try
							{
								l_e = Double.parseDouble(str);
								if (l_e < 0) throw new RuntimeException();

								break;
							}
							catch (Throwable ex)
							{
								ErrorHandling.showErrorDialog("Non-valid link length value. Please, introduce a non-negative number", "Error setting link length value");
							}
						}

						NetPlan netPlan = callback.getDesign();
						
						try
						{
							Set<Long> linkIds = netPlan.getLinkIds();
							for (long linkId : linkIds) netPlan.setLinkLengthInKm(linkId, l_e);

							callback.updateNetPlanView();
						}
						catch (Exception ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to set link length value to all links");
						}
					}
				});

				options.add(lengthToAll);

				JMenuItem lengthToAllEuclidean = new JMenuItem("Set (all) link length to node-pair Euclidean distance");

				lengthToAllEuclidean.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						NetPlan netPlan = callback.getDesign();
						
						try
						{
							Set<Long> linkIds = netPlan.getLinkIds();
							for (long linkId : linkIds)
							{
								long originNodeId = netPlan.getLinkOriginNode(linkId);
								long destinationNodeId = netPlan.getLinkDestinationNode(linkId);
								double euclideanDistance = netPlan.getNodePairEuclideanDistance(originNodeId, destinationNodeId);
								netPlan.setLinkLengthInKm(linkId, euclideanDistance);
							}

							callback.updateNetPlanView();
						}
						catch (Exception ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to set link length value to all links");
						}
					}
				});

				options.add(lengthToAllEuclidean);

				JMenuItem lengthToAllHaversine = new JMenuItem("Set (all) link length to node-pair Haversine distance (longitude-latitude)");

				lengthToAllHaversine.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						NetPlan netPlan = callback.getDesign();
						
						try
						{
							Set<Long> linkIds = netPlan.getLinkIds();
							for (long linkId : linkIds)
							{
								long originNodeId = netPlan.getLinkOriginNode(linkId);
								long destinationNodeId = netPlan.getLinkDestinationNode(linkId);
								double haversineDistanceInKm = netPlan.getNodePairHaversineDistanceInKm(originNodeId, destinationNodeId);
								netPlan.setLinkLengthInKm(linkId, haversineDistanceInKm);
							}

							callback.updateNetPlanView();
						}
						catch (Exception ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to set link length value to all links");
						}
					}
				});

				options.add(lengthToAllHaversine);
				
				JMenuItem scaleLinkLength = new JMenuItem("Scale (all) link length");

				scaleLinkLength.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						double scaleFactor;

						while (true)
						{
							String str = JOptionPane.showInputDialog(null, "(Multiplicative) Scale factor", "Scale (all) link length", JOptionPane.QUESTION_MESSAGE);
							if (str == null) return;

							try
							{
								scaleFactor = Double.parseDouble(str);
								if (scaleFactor < 0) throw new RuntimeException();

								break;
							}
							catch (Throwable ex)
							{
								ErrorHandling.showErrorDialog("Non-valid scale value. Please, introduce a non-negative number", "Error setting scale factor");
							}
						}

						NetPlan netPlan = callback.getDesign();
						
						try
						{
							Set<Long> linkIds = netPlan.getLinkIds();
							for (long linkId : linkIds) netPlan.setLinkLengthInKm(linkId, netPlan.getLinkLengthInKm(linkId) * scaleFactor);

							callback.updateNetPlanView();
						}
						catch (Exception ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to scale link length");
						}
					}
				});

				options.add(scaleLinkLength);

				if (itemId != null && netPlan.isMultiLayer())
				{
					final long linkId = (long) itemId;
					if (netPlan.isLinkCoupled(linkId))
					{
						JMenuItem decoupleLinkItem = new JMenuItem("Decouple link");
						decoupleLinkItem.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								netPlan.decoupleLink(linkId);
								model.setValueAt("", row, 5);
								callback.updateWarnings();
							}
						});

						options.add(decoupleLinkItem);
					}
					else
					{
						JMenuItem createLowerLayerDemandFromLinkItem = new JMenuItem("Create lower layer demand from link");
						createLowerLayerDemandFromLinkItem.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								Set<Long> layerIds = netPlan.getLayerIds();
								final JComboBox layerSelector = new WiderJComboBox();
								for (long layerId : layerIds)
								{
									if (layerId == netPlan.getLayerDefaultId()) continue;
									
									final String layerName = netPlan.getLayerName(layerId);
									String layerLabel = "Layer " + layerId;
									if (!layerName.isEmpty()) layerLabel += " (" + layerName + ")";

									layerSelector.addItem(StringLabeller.of(layerId, layerLabel));
								}

								layerSelector.setSelectedIndex(0);

								JPanel pane = new JPanel();
								pane.add(new JLabel("Select layer: "));
								pane.add(layerSelector);

								while (true)
								{
									int result = JOptionPane.showConfirmDialog(null, pane, "Please select the lower layer to create the demand", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
									if (result != JOptionPane.OK_OPTION) return;

									try
									{
										long layerId = (long) ((StringLabeller) layerSelector.getSelectedItem()).getObject();
										netPlan.createLowerLayerDemandFromLink(linkId, layerId);

										callback.updateNetPlanView();
										break;
									}
									catch (Throwable ex)
									{
										ErrorHandling.showErrorDialog(ex.getMessage(), "Error creating lower layer demand from link");
									}
								}
							}
						});

						options.add(createLowerLayerDemandFromLinkItem);

						JMenuItem coupleLinkToDemand = new JMenuItem("Couple link to lower layer demand");
						coupleLinkToDemand.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								Set<Long> layerIds = netPlan.getLayerIds();
								final JComboBox layerSelector = new WiderJComboBox();
								final JComboBox demandSelector = new WiderJComboBox();
								for (long layerId : layerIds)
								{
									if (layerId == netPlan.getLayerDefaultId()) continue;
									
									final String layerName = netPlan.getLayerName(layerId);
									String layerLabel = "Layer " + layerId;
									if (!layerName.isEmpty()) layerLabel += " (" + layerName + ")";

									layerSelector.addItem(StringLabeller.of(layerId, layerLabel));
								}
								
								layerSelector.addItemListener(new ItemListener()
								{
									@Override
									public void itemStateChanged(ItemEvent e)
									{
										if (layerSelector.getSelectedIndex() >= 0)
										{
											long selectedLayerId = (Long) ((StringLabeller) layerSelector.getSelectedItem()).getObject();

											demandSelector.removeAllItems();
											Set<Long> demandIds_thisLayer = netPlan.getDemandIds(selectedLayerId);
											for(long demandId : demandIds_thisLayer)
											{
												if (netPlan.isDemandCoupled(selectedLayerId, demandId)) continue;

												long ingressNodeId = netPlan.getDemandIngressNode(selectedLayerId, demandId);
												long egressNodeId = netPlan.getDemandEgressNode(selectedLayerId, demandId);
												String ingressNodeName = netPlan.getNodeName(ingressNodeId);
												String egressNodeName = netPlan.getNodeName(egressNodeId);

												demandSelector.addItem(new StringLabeller(demandId, "d" + demandId + " [n" + ingressNodeId + " (" + ingressNodeName + ") -> n" + egressNodeId + " (" + egressNodeName + ")]"));
											}
										}
										
										if (demandSelector.getItemCount() == 0)
										{
											demandSelector.setEnabled(false);
										}
										else 
										{
											demandSelector.setSelectedIndex(0);
											demandSelector.setEnabled(true);
										}
									}
								});

								layerSelector.setSelectedIndex(-1);
								layerSelector.setSelectedIndex(0);

								JPanel pane = new JPanel(new MigLayout("", "[][grow]", "[][]"));
								pane.add(new JLabel("Select layer: "));
								pane.add(layerSelector, "growx, wrap");
								pane.add(new JLabel("Select demand: "));
								pane.add(demandSelector, "growx, wrap");

								while (true)
								{
									int result = JOptionPane.showConfirmDialog(null, pane, "Please select the lower layer demand", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
									if (result != JOptionPane.OK_OPTION) return;

									try
									{
										long layerId = (long) ((StringLabeller) layerSelector.getSelectedItem()).getObject();
										long demandId;
										try { demandId = (long) ((StringLabeller) demandSelector.getSelectedItem()).getObject(); }
										catch(Throwable ex) { throw new RuntimeException("No demand was selected"); }
										
										netPlan.coupleLinkWithLowerLayerDemand(linkId, layerId, demandId);

										callback.updateNetPlanView();
										break;
									}
									catch (Throwable ex)
									{
										ErrorHandling.showErrorDialog(ex.getMessage(), "Error coupling lower layer demand to link");
									}
								}
							}
						});

						options.add(coupleLinkToDemand);
					}

					final Set<Long> coupledLinks = netPlan.getLinksCoupledDemandMap().keySet();
					if (!coupledLinks.isEmpty())
					{
						JMenuItem decoupleAllLinksItem = new JMenuItem("Decouple all links");
						decoupleAllLinksItem.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								for(long linkId : new LinkedHashSet<Long>(coupledLinks)) netPlan.decoupleLink(linkId);

								int numRows = model.getRowCount();
								for(int i = 0; i < numRows; i++) model.setValueAt("", i, 5);

								callback.updateWarnings();
							}
						});

						options.add(decoupleAllLinksItem);
					}
					
					if (coupledLinks.size() < netPlan.getNumberOfLinks())
					{
						JMenuItem createLowerLayerDemandsFromLinksItem = new JMenuItem("Create lower layer demands from uncoupled links");
						createLowerLayerDemandsFromLinksItem.addActionListener(new ActionListener()
						{
							@Override
							public void actionPerformed(ActionEvent e)
							{
								Set<Long> layerIds = netPlan.getLayerIds();
								final JComboBox layerSelector = new WiderJComboBox();
								for (long layerId : layerIds)
								{
									if (layerId == netPlan.getLayerDefaultId()) continue;
									
									final String layerName = netPlan.getLayerName(layerId);
									String layerLabel = "Layer " + layerId;
									if (!layerName.isEmpty()) layerLabel += " (" + layerName + ")";

									layerSelector.addItem(StringLabeller.of(layerId, layerLabel));
								}

								layerSelector.setSelectedIndex(0);

								JPanel pane = new JPanel();
								pane.add(new JLabel("Select layer: "));
								pane.add(layerSelector);

								while (true)
								{
									int result = JOptionPane.showConfirmDialog(null, pane, "Please select the lower layer to create demands", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
									if (result != JOptionPane.OK_OPTION) return;

									try
									{
										long layerId = (long) ((StringLabeller) layerSelector.getSelectedItem()).getObject();
										netPlan.createLowerLayerDemandsFromLinks(layerId);

										callback.updateNetPlanView();
										break;
									}
									catch (Throwable ex)
									{
										ErrorHandling.showErrorDialog(ex.getMessage(), "Error creating lower layer demands");
									}
								}
							}
						});

						options.add(createLowerLayerDemandsFromLinksItem);
					}
				}
				
				break;

			default:
				throw new RuntimeException("Bad");
		}

		return options;
	}

	private List<JMenuItem> getForcedOptions()
	{
		List<JMenuItem> options = new LinkedList<JMenuItem>();

		switch (networkElementType)
		{
			case NODE:
				JMenuItem showAllNodes = new JMenuItem("Show all nodes");
				showAllNodes.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						int numRows = model.getRowCount();
						for(int row = 0; row < numRows; row++)
							if (model.getValueAt(row, 1) != null)
								model.setValueAt(true, row, 1);
					}
				});

				options.add(showAllNodes);

				JMenuItem hideAllNodes = new JMenuItem("Hide all nodes");
				hideAllNodes.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						int numRows = model.getRowCount();
						for(int row = 0; row < numRows; row++)
							if (model.getValueAt(row, 1) != null)
								model.setValueAt(false, row, 1);
					}
				});

				options.add(hideAllNodes);
				
				break;

			case LINK:
				JMenuItem showAllLinks = new JMenuItem("Show all links");
				showAllLinks.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						int numRows = model.getRowCount();
						for(int row = 0; row < numRows; row++)
							if (model.getValueAt(row, 1) != null)
								model.setValueAt(true, row, 1);
					}
				});

				options.add(showAllLinks);

				JMenuItem hideAllLinks = new JMenuItem("Hide all links");
				hideAllLinks.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						int numRows = model.getRowCount();
						for(int row = 0; row < numRows; row++)
							if (model.getValueAt(row, 1) != null)
								model.setValueAt(false, row, 1);
					}
				});

				options.add(hideAllLinks);
				
				break;
				
			default:
				break;
		}

		return options;
	}

	private List<JMenuItem> getExtraAddOptions()
	{
		List<JMenuItem> options = new LinkedList<JMenuItem>();
		NetPlan netPlan = callback.getDesign();

		switch (networkElementType)
		{
			case LAYER:
			case NODE:
			case DEMAND:
			case ROUTE:
			case PROTECTION_SEGMENT:
				break;
				
			case LINK:
				final JMenuItem fullMeshEuclidean = new JMenuItem("Generate full-mesh (link length as Euclidean distance)");
				final JMenuItem fullMeshHaversine = new JMenuItem("Generate full-mesh (link length as Haversine distance)");
				options.add(fullMeshEuclidean);
				options.add(fullMeshHaversine);

				fullMeshEuclidean.addActionListener(new FullMeshActionListener(true));
				fullMeshHaversine.addActionListener(new FullMeshActionListener(false));

				if (netPlan.getNumberOfNodes() < 2)
				{
					fullMeshEuclidean.setEnabled(false);
					fullMeshHaversine.setEnabled(false);
				}
				break;
				
			case FORWARDING_RULE:
				final JMenuItem ecmpRouting = new JMenuItem("Generate ECMP forwarding rules from link IGP weights");
				options.add(ecmpRouting);
				
				ecmpRouting.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						NetPlan netPlan = callback.getDesign();
						Map<Long, Double> linkWeightMap = IPUtils.getLinkWeightMap(netPlan);
						IPUtils.setECMPForwardingRulesFromLinkWeights(netPlan, linkWeightMap);
						callback.updateNetPlanView();
					}
				});

				if (!netPlan.hasLinks() || !netPlan.hasDemands()) ecmpRouting.setEnabled(false);
				break;

			case SRG:

				JMenu submenuSRGs = new JMenu("Add SRGs from model");
				options.add(submenuSRGs);

				final JMenuItem onePerNode = new JMenuItem("One SRG per node");
				submenuSRGs.add(onePerNode);
				final JMenuItem onePerLink = new JMenuItem("One SRG per unidirectional link");
				submenuSRGs.add(onePerLink);
				final JMenuItem onePerLinkBundle = new JMenuItem("One SRG per unidirectional bundle of links");
				submenuSRGs.add(onePerLinkBundle);
				final JMenuItem onePerBidiLinkBundle = new JMenuItem("One SRG per bidirectional bundle of links");
				submenuSRGs.add(onePerBidiLinkBundle);
				
				if (!netPlan.hasNodes())
				{
					onePerNode.setEnabled(false);
				}

				if (!netPlan.hasLinks())
				{
					onePerLink.setEnabled(false);
					onePerLinkBundle.setEnabled(false);
					onePerBidiLinkBundle.setEnabled(false);
				}
				
				ActionListener srgModel = new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						double mttf;
						double mttr;

						JTextField txt_mttf = new JTextField(5);
						JTextField txt_mttr = new JTextField(5);
						JCheckBox chk_removeExistingSRGs = new JCheckBox();
						chk_removeExistingSRGs.setSelected(true);

						JPanel pane = new JPanel(new GridLayout(0, 2));
						pane.add(new JLabel("MTTF (in hours, zero or negative value means no failure): "));
						pane.add(txt_mttf);
						pane.add(new JLabel("MTTR (in hours): "));
						pane.add(txt_mttr);
						pane.add(new JLabel("Remove existing SRGs: "));
						pane.add(chk_removeExistingSRGs);

						while (true)
						{
							int result = JOptionPane.showConfirmDialog(null, pane, "Add SRGs from model", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
							if (result != JOptionPane.OK_OPTION)
							{
								return;
							}

							try
							{
								mttf = Double.parseDouble(txt_mttf.getText());
								mttr = Double.parseDouble(txt_mttr.getText());
								if (mttr <= 0) throw new IllegalArgumentException("MTTR must be greater than zero");

								break;
							}
							catch (Throwable ex)
							{
								ErrorHandling.showErrorDialog(ex.getMessage(), "Error adding SRGs from model");
							}
						}

						boolean removeExistingSRGs = chk_removeExistingSRGs.isSelected();

						NetPlan netPlan = callback.getDesign();

						try
						{
							SharedRiskModel srgModel = null;
							if (e.getSource() == onePerNode) srgModel = SRGUtils.SharedRiskModel.PER_NODE;
							else if (e.getSource() == onePerLink) srgModel = SRGUtils.SharedRiskModel.PER_LINK;
							else if (e.getSource() == onePerLinkBundle) srgModel = SRGUtils.SharedRiskModel.PER_DIRECTIONAL_LINK_BUNDLE;
							else if (e.getSource() == onePerBidiLinkBundle) srgModel = SRGUtils.SharedRiskModel.PER_BIDIRECTIONAL_LINK_BUNDLE;

							if (srgModel == null) throw new RuntimeException("Bad");
							SRGUtils.configureSRGs(netPlan, mttf, mttr, srgModel, removeExistingSRGs);

							callback.updateNetPlanView();
						}
						catch (Exception ex)
						{
							ErrorHandling.showErrorDialog(ex.getMessage(), "Unable to add SRGs from model");
						}

					}
				};

				onePerNode.addActionListener(srgModel);
				onePerLink.addActionListener(srgModel);
				onePerLinkBundle.addActionListener(srgModel);
				onePerBidiLinkBundle.addActionListener(srgModel);

				break;

			default:
				throw new RuntimeException("Bad");
		}

		return options;
	}
	
	private static void setMaxSize(JComponent c)
	{
		final Dimension max = c.getMaximumSize();
		final Dimension pref = c.getPreferredSize();

		max.height = pref.height;
		c.setMaximumSize(max);
	}

	private static void viewEditBackupSegmentListGUI(final INetworkCallback callback, final TopologyPanel topologyPanel, final long routeId)
	{
		final NetPlan netPlan = callback.getDesign();

		Set<Long> candidateSegmentIds = new TreeSet<Long>();
		Set<Long> currentSegmentIds = new TreeSet<Long>();

		final List<Long> seqLinks = netPlan.getRouteSequenceOfLinks(routeId);

		Set<Long> segmentIds = netPlan.getProtectionSegmentIds();
		for (long segmentId : segmentIds)
		{
			try
			{
				/* Only considered a segment as candidate if it is mergeable to the current route */
				List<Long> seqLinks_thisSegment = netPlan.getProtectionSegmentSequenceOfLinks(segmentId);
				List<Long> mergedRoute = netPlan.getMergedPath(seqLinks, seqLinks_thisSegment); /* Not remove! It's used to check mergeability */
				candidateSegmentIds.add(segmentId);
			}
			catch(Throwable e)
			{
			}
		}

		if (candidateSegmentIds.isEmpty()) throw new Net2PlanException("No segment can be applied to this route");

		currentSegmentIds.addAll(netPlan.getRouteBackupSegmentList(routeId));
		candidateSegmentIds.removeAll(currentSegmentIds);

		final JComboBox segmentSelector = new WiderJComboBox();

		final DefaultTableModel model = new ClassAwareTableModel(new Object[1][6], new String[] { "Id", "Seq. links", "Seq. nodes", "Reserved capacity", "", "" })
		{
			@Override
			public boolean isCellEditable(int rowIndex, int columnIndex)
			{
				return columnIndex == 4 || columnIndex == 5;
			}
		};
		final JTable table = new AdvancedJTable(model);
		table.setEnabled(false);

		final JPanel addSegment_pnl = new JPanel(new MigLayout("", "[grow][][]", "[]"));
		JButton addSegment_btn = new JButton("Add");
		addSegment_btn.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				Object selectedItem = segmentSelector.getSelectedItem();
				long segmentId = (Long) ((StringLabeller) selectedItem).getObject();
				netPlan.addProtectionSegmentToRouteBackupSegmentList(segmentId, routeId);
				callback.updateNetPlanView();

				segmentSelector.removeItem(selectedItem);
				if (segmentSelector.getItemCount() == 0) addSegment_pnl.setVisible(false);

				List<Long> segmentSeqLinks = netPlan.getProtectionSegmentSequenceOfLinks(segmentId);
				List<Long> segmentSeqNodes = netPlan.getProtectionSegmentSequenceOfNodes(segmentId);
				double reservedCapacity = netPlan.getProtectionSegmentReservedCapacity(segmentId);

				if (!table.isEnabled()) model.removeRow(0);
				model.addRow(new Object[] { segmentId, CollectionUtils.join(segmentSeqLinks, " => "), CollectionUtils.join(segmentSeqNodes, " => "), reservedCapacity, "Remove", "View" });

				table.setEnabled(true);
			}
		});

		JButton viewSegment_btn1 = new JButton("View");
		viewSegment_btn1.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				Object selectedItem = segmentSelector.getSelectedItem();
				long segmentId = (Long) ((StringLabeller) selectedItem).getObject();

				List<Long> segmentSeqLinks = netPlan.getProtectionSegmentSequenceOfLinks(segmentId);
				topologyPanel.getCanvas().showRoutes(seqLinks, segmentSeqLinks);
			}
		});

		addSegment_pnl.add(segmentSelector, "growx, wmin 50");
		addSegment_pnl.add(addSegment_btn);
		addSegment_pnl.add(viewSegment_btn1);

		Action delete = new AbstractAction()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				try
				{
					JTable table = (JTable) e.getSource();
					int modelRow = Integer.parseInt(e.getActionCommand());

					long segmentId = (Long) table.getModel().getValueAt(modelRow, 0);
					netPlan.removeProtectionSegmentFromRouteBackupSegmentList(segmentId, routeId);
					callback.updateNetPlanView();

					List<Long> segmentSeqLinks = netPlan.getProtectionSegmentSequenceOfLinks(segmentId);
					List<Long> segmentSeqNodes = netPlan.getProtectionSegmentSequenceOfNodes(segmentId);
					double reservedCapacity = netPlan.getProtectionSegmentReservedCapacity(segmentId);

					String segmentLabel = "Protection segment " + segmentId;
					segmentLabel += ": seq. links = " + CollectionUtils.join(segmentSeqLinks, " => ");
					segmentLabel += ", seq. nodes = " + CollectionUtils.join(segmentSeqNodes, " => ");
					segmentLabel += ", reserved capacity = " + reservedCapacity;

					segmentSelector.addItem(StringLabeller.of(segmentId, segmentLabel));

					((DefaultTableModel) table.getModel()).removeRow(modelRow);

					table.setEnabled(true);

					if (table.getModel().getRowCount() == 0)
					{
						((DefaultTableModel) table.getModel()).addRow(new Object[6]);
						table.setEnabled(false);
					}
				}
				catch (Throwable e1)
				{
				}
			}
		};

		Action view = new AbstractAction()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				try
				{
					JTable table = (JTable) e.getSource();
					int modelRow = Integer.parseInt(e.getActionCommand());

					long segmentId = (Long) table.getModel().getValueAt(modelRow, 0);
					List<Long> segmentSeqLinks = netPlan.getProtectionSegmentSequenceOfLinks(segmentId);
					topologyPanel.getCanvas().showRoutes(seqLinks, segmentSeqLinks);
				}
				catch (Throwable e1)
				{
				}
			}
		};

		new ButtonColumn(table, delete, 4);
		new ButtonColumn(table, view, 5);

		final JScrollPane scrollPane = new JScrollPane(table, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		scrollPane.setBorder(BorderFactory.createTitledBorder(new LineBorder(Color.BLACK), "Current backup segment list"));
		scrollPane.setAlignmentY(JScrollPane.TOP_ALIGNMENT);

		final JDialog dialog = new JDialog();
		dialog.setLayout(new BorderLayout());
		dialog.add(addSegment_pnl, BorderLayout.NORTH);
		dialog.add(scrollPane, BorderLayout.CENTER);

		for (long segmentId : candidateSegmentIds)
		{
			List<Long> segmentSeqLinks = netPlan.getProtectionSegmentSequenceOfLinks(segmentId);
			List<Long> segmentSeqNodes = netPlan.getProtectionSegmentSequenceOfNodes(segmentId);
			double reservedCapacity = netPlan.getProtectionSegmentReservedCapacity(segmentId);

			String segmentLabel = "Protection segment " + segmentId;
			segmentLabel += ": seq. links = " + CollectionUtils.join(segmentSeqLinks, " => ");
			segmentLabel += ", seq. nodes = " + CollectionUtils.join(segmentSeqNodes, " => ");
			segmentLabel += ", reserved capacity = " + reservedCapacity;

			segmentSelector.addItem(StringLabeller.of(segmentId, segmentLabel));
		}

		if (segmentSelector.getItemCount() == 0)
		{
			addSegment_pnl.setVisible(false);
		}
		else
		{
			segmentSelector.setSelectedIndex(0);
		}

		if (!currentSegmentIds.isEmpty())
		{
			model.removeRow(0);

			for (long segmentId : currentSegmentIds)
			{
				List<Long> segmentSeqLinks = netPlan.getProtectionSegmentSequenceOfLinks(segmentId);
				List<Long> segmentSeqNodes = netPlan.getProtectionSegmentSequenceOfNodes(segmentId);
				double reservedCapacity = netPlan.getProtectionSegmentReservedCapacity(segmentId);

				model.addRow(new Object[] { segmentId, CollectionUtils.join(segmentSeqLinks, " => "), CollectionUtils.join(segmentSeqNodes, " => "), reservedCapacity, "Remove", "View" });
			}

			table.setEnabled(true);
		}

		table.setDefaultRenderer(Boolean.class, new UnfocusableCellRenderer());
		table.setDefaultRenderer(Double.class, new UnfocusableCellRenderer());
		table.setDefaultRenderer(Object.class, new UnfocusableCellRenderer());
		table.setDefaultRenderer(Float.class, new UnfocusableCellRenderer());
		table.setDefaultRenderer(Long.class, new UnfocusableCellRenderer());
		table.setDefaultRenderer(Integer.class, new UnfocusableCellRenderer());
		table.setDefaultRenderer(String.class, new UnfocusableCellRenderer());

		double x_p = netPlan.getRouteCarriedTraffic(routeId);
		double y_p = netPlan.getRouteOccupiedCapacity(routeId);
		dialog.setTitle("View/edit backup segment list for route " + routeId + " (carried traffic = " + x_p + ", occupied capacity = " + y_p + ")");
		SwingUtils.configureCloseDialogOnEscape(dialog);
		dialog.setModalityType(Dialog.ModalityType.APPLICATION_MODAL);
		dialog.setSize(new Dimension(500, 300));
		dialog.setLocationRelativeTo(null);
		dialog.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		dialog.setVisible(true);

		topologyPanel.getCanvas().resetPickedState();
	}

	private class FullMeshActionListener implements ActionListener
	{
		private final boolean euclidean;

		public FullMeshActionListener(boolean euclidean)
		{
			this.euclidean = euclidean;
		}

		@Override
		public void actionPerformed(ActionEvent e)
		{
			NetPlan netPlan = callback.getDesign();
			netPlan.removeAllLinks();
			for(long nodeId_1 : netPlan.getNodeIds())
			{
				for(long nodeId_2 : netPlan.getNodeIds())
				{
					if (nodeId_1 >= nodeId_2) continue;
					
					Pair<Long, Long> out = netPlan.addLinkBidirectional(nodeId_1, nodeId_2, 0, euclidean ? netPlan.getNodePairEuclideanDistance(nodeId_1, nodeId_2) : netPlan.getNodePairHaversineDistanceInKm(nodeId_1, nodeId_2), null);
					topologyPanel.getCanvas().addLink(out.getFirst(), nodeId_1, nodeId_2);
					topologyPanel.getCanvas().addLink(out.getSecond(), nodeId_2, nodeId_1);
				}
			}
			
			topologyPanel.getCanvas().refresh();
			callback.updateNetPlanView();
		}
	}
}