/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.gui.utils.topology;

import com.net2plan.interfaces.networkDesign.NetPlan;
import java.awt.geom.Point2D;

/**
 * Interface to be implemented by any class dealing with network topologies.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.0
 */
public interface ITopologyCallback
{
	/**
	 * Adds a node at the given coordinates.
	 * 
	 * @param pos 2D position
	 * @since 0.2.0
	 */
	public void addNode(Point2D pos);

	/**
	 * It is called when a new network design is loaded.
	 * 
	 * @param netPlan Network design
	 * @since 0.2.0
	 */
	public void loadDesign(NetPlan netPlan);

	/**
	 * Returns the current network design.
	 *
	 * @return Current <code>NetPlan</code>
	 * @since 0.3.0
	 */
	public NetPlan getDesign();
	
	/**
	 * Indicates whether or not the design is editable after loading.
	 *
	 * @return {@code true} if it is editable, and {@code false} otherwise
	 * @since 0.3.0
	 */
	public boolean isEditable();

	/**
	 * Loads a set of traffic demands from the given <code>NetPlan</code>.
	 *
	 * @param netPlan Network design containing a demand set
	 * @since 0.2.3
	 */
	public void loadTrafficDemands(NetPlan netPlan);

	/**
	 * Moves the node to the given position.
	 * 
	 * @param nodeId Node identifier
	 * @param pos 2D position
	 * @since 0.3.0
	 */
	public void moveNode(long nodeId, Point2D pos);

	/**
	 * Removes the given node.
	 * 
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public void removeNode(long nodeId);

	/**
	 * Removes the given link.
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void removeLink(long linkId);

	/**
	 * Adds a new link in the current layer.
	 * 
	 * @param originNodeId Origin node id
	 * @param destinationNodeId Destination node id
	 * @since 0.3.0
	 */
	public void addLink(long originNodeId, long destinationNodeId);

	/**
	 * Shows the given node.
	 * 
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public void showNode(long nodeId);

	/**
	 * Shows the given link.
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public void showLink(long linkId);

	/**
	 * Resets the current topology (i.e. remove any node/link).
	 * 
	 * @since 0.3.0
	 */
	public void reset();

	/**
	 * Resets the current view (i.e. reset picked state).
	 * 
	 * @since 0.3.0
	 */
	public void resetView();

}
