/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.gui.utils.topology;

import com.net2plan.internal.plugins.ITopologyCanvas;
import com.net2plan.gui.utils.FileDrop;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.internal.ErrorHandling;
import com.net2plan.internal.SystemUtils;
import com.net2plan.gui.utils.FileChooserNetworkDesign;
import com.net2plan.gui.utils.SwingUtils;
import com.net2plan.internal.Constants.DialogType;
import java.awt.Color;
import java.awt.Component;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.io.File;
import java.util.List;
import java.util.Locale;
import javax.swing.AbstractButton;
import javax.swing.Box;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;
import javax.swing.border.LineBorder;
import net.miginfocom.swing.MigLayout;

/**
 * Wrapper class for the graph canvas.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.0
 */
public class TopologyPanel extends JPanel implements ActionListener
{
	private ITopologyCanvas canvas;
	private JButton btn_load, btn_loadDemand, btn_save, btn_zoomIn, btn_zoomOut, btn_zoomAll, btn_takeSnapshot, btn_reset;
	private JToggleButton btn_showNodeNames, btn_showLinkIds, btn_showNonConnectedNodes;
	private FileChooserNetworkDesign fc_netPlan;
	private JFileChooser fc_demands;
	private final ITopologyCallback callback;
	private final File defaultDesignDirectory, defaultDemandDirectory;
	private final JLabel position;
	
	
	/**
	 * Simplified constructor that does not require to indicate default locations 
	 * for <code>.n2p</code> files.
	 * 
	 * @param callback Topology callback listening plugin events
	 * @param canvasType Canvas type (i.e. JUNG)
	 * @since 0.2.3
	 */
	public TopologyPanel(ITopologyCallback callback, Class<? extends ITopologyCanvas> canvasType)
	{
		this(callback, canvasType, null);
	}
	
	/**
	 * Simplified constructor that does not require to indicate default locations 
	 * for <code>.n2p</code> files.
	 * 
	 * @param callback Topology callback listening plugin events
	 * @param canvasType Canvas type (i.e. JUNG)
	 * @param plugins List of plugins to be included (it may be null)
	 * @since 0.2.3
	 */
	public TopologyPanel(ITopologyCallback callback, Class<? extends ITopologyCanvas> canvasType, List<ITopologyCanvasPlugin> plugins)
	{
		this(callback, null, null, canvasType, plugins);
	}

	/**
	 * Default constructor.
	 * 
	 * @param callback Topology callback listening plugin events
	 * @param defaultDesignDirectory Default location for design <code>.n2p</code> files (it may be null, then default is equal to <code>net2planFolder/workspace/data/networkTopologies</code>)
	 * @param defaultDemandDirectory Default location for design <code>.n2p</code> files (it may be null, then default is equal to <code>net2planFolder/workspace/data/trafficMatrices</code>)
	 * @param canvasType Canvas type (i.e. JUNG)
	 * @param plugins List of plugins to be included (it may be null)
	 * @since 0.2.0
	 */
	public TopologyPanel(final ITopologyCallback callback, File defaultDesignDirectory, File defaultDemandDirectory, Class<? extends ITopologyCanvas> canvasType, List<ITopologyCanvasPlugin> plugins)
	{
		File currentDir = SystemUtils.getCurrentDir();

		this.callback = callback;
		this.defaultDesignDirectory = defaultDesignDirectory == null ? new File(currentDir + SystemUtils.getDirectorySeparator() + "workspace" + SystemUtils.getDirectorySeparator() + "data" + SystemUtils.getDirectorySeparator() + "networkTopologies") : defaultDesignDirectory;
		this.defaultDemandDirectory = defaultDemandDirectory == null ? new File(currentDir + SystemUtils.getDirectorySeparator() + "workspace" + SystemUtils.getDirectorySeparator() + "data" + SystemUtils.getDirectorySeparator() + "trafficMatrices") : defaultDemandDirectory;

		try { canvas = canvasType.newInstance(); }
		catch (InstantiationException | IllegalAccessException e) { throw new RuntimeException(e); }

		if (plugins != null)
			for (ITopologyCanvasPlugin plugin : plugins)
				addPlugin(plugin);

		setLayout(new MigLayout("insets 0 0 0 0", "[grow]", "[][grow]"));

		JToolBar toolbar = new JToolBar();
		toolbar.setRollover(true);
		toolbar.setFloatable(false);
		toolbar.setOpaque(false);
		toolbar.setBorderPainted(false);
		add(toolbar, "growx, wrap");

		JComponent canvasComponent = canvas.getComponent();
		canvasComponent.setBorder(LineBorder.createBlackLineBorder());
		add(canvasComponent, "grow");

		btn_load = new JButton();
		btn_load.setToolTipText("Load a network design");
		btn_loadDemand = new JButton();
		btn_loadDemand.setToolTipText("Load a traffic demand set");
		btn_save = new JButton();
		btn_save.setToolTipText("Save current state to a file");
		btn_zoomIn = new JButton();
		btn_zoomIn.setToolTipText("Zoom in");
		btn_zoomOut = new JButton();
		btn_zoomOut.setToolTipText("Zoom out");
		btn_zoomAll = new JButton();
		btn_zoomAll.setToolTipText("Zoom all"); 
		btn_takeSnapshot = new JButton();
		btn_takeSnapshot.setToolTipText("Take a snapshot of the canvas");
		btn_showNodeNames = new JToggleButton("Node names"); btn_showNodeNames.setToolTipText("Show/hide node names");
		btn_showLinkIds = new JToggleButton("Link ids"); btn_showLinkIds.setToolTipText("Show/hide link identifiers");
		btn_showNonConnectedNodes = new JToggleButton("Non-connected nodes"); btn_showNonConnectedNodes.setToolTipText("Show/hide non-connected nodes");
		JButton increaseNodeSize = new JButton();
		increaseNodeSize.setToolTipText("Increase node size"); 
		JButton decreaseNodeSize = new JButton();
		decreaseNodeSize.setToolTipText("Decrease node size");
		JButton increaseFontSize = new JButton();
		increaseFontSize.setToolTipText("Increase font size");
		JButton decreaseFontSize = new JButton();
		decreaseFontSize.setToolTipText("Decrease font size");
		
		btn_reset = new JButton("Reset");
		btn_reset.setToolTipText("Reset the user interface");

		btn_load.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/loadDesign.png")));
		btn_loadDemand.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/loadDemand.png")));
		btn_save.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/saveDesign.png")));
		btn_showNodeNames.setBorderPainted(true);
		btn_showLinkIds.setBorderPainted(true);
		btn_showNonConnectedNodes.setBorderPainted(true);
		btn_zoomIn.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/zoomIn.png")));
		btn_zoomOut.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/zoomOut.png")));
		btn_zoomAll.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/zoomAll.png")));
		btn_takeSnapshot.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/takeSnapshot.png")));
		increaseNodeSize.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/increaseNode.png")));
		decreaseNodeSize.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/decreaseNode.png")));
		increaseFontSize.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/increaseFont.png")));
		decreaseFontSize.setIcon(new ImageIcon(TopologyPanel.class.getResource("/resources/gui/decreaseFont.png")));

		btn_load.addActionListener(this);
		btn_loadDemand.addActionListener(this);
		btn_save.addActionListener(this);
		btn_showNodeNames.addActionListener(this);
		btn_showLinkIds.addActionListener(this);
		btn_showNonConnectedNodes.addActionListener(this);
		btn_zoomIn.addActionListener(this);
		btn_zoomOut.addActionListener(this);
		btn_zoomAll.addActionListener(this);
		btn_takeSnapshot.addActionListener(this);
		btn_reset.addActionListener(this);

		toolbar.add(btn_load);
		toolbar.add(btn_loadDemand);
		toolbar.add(btn_save);
		toolbar.add(new JToolBar.Separator());
		toolbar.add(btn_zoomIn);
		toolbar.add(btn_zoomOut);
		toolbar.add(btn_zoomAll);
		toolbar.add(btn_takeSnapshot);
		toolbar.add(new JToolBar.Separator());
		toolbar.add(btn_showNodeNames);
		toolbar.add(btn_showLinkIds);
		toolbar.add(btn_showNonConnectedNodes);
		toolbar.add(increaseNodeSize);
		toolbar.add(decreaseNodeSize);
		toolbar.add(increaseFontSize);
		toolbar.add(decreaseFontSize);
		toolbar.add(Box.createHorizontalGlue());
		toolbar.add(btn_reset);

		increaseNodeSize.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				getCanvas().increaseNodeSize();
			}
		});
		
		decreaseNodeSize.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				getCanvas().decreaseNodeSize();
			}
		});
		
		increaseFontSize.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				getCanvas().increaseFontSize();
			}
		});
		
		decreaseFontSize.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				getCanvas().decreaseFontSize();
			}
		});
		
		List<Component> children = SwingUtils.getAllComponents(this);
		for(Component component : children)
			if (component instanceof AbstractButton)
				component.setFocusable(false);
		
		if (ErrorHandling.isDebugEnabled())
		{
			canvasComponent.addMouseMotionListener(new MouseMotionListener()
			{
				@Override
				public void mouseDragged(MouseEvent e) { }

				@Override
				public void mouseMoved(MouseEvent e)
				{
					Point point = e.getPoint();
					position.setText("view = " + point + ", layout = " + canvas.convertViewCoordinatesToRealCoordinates(point));
				}
			});
			
			position = new JLabel();
			add(position, "dock south, growx");
		}
		else
		{
			position = null;
		}
		
		new FileDrop(canvasComponent, new LineBorder(Color.BLACK), new FileDrop.Listener()
        {
			@Override
			public void filesDropped(File[] files)
            {  
				for(File file : files)
                {
					try
                    {
						if (!file.getName().toLowerCase(Locale.getDefault()).endsWith(".n2p")) return;
						loadDesignFromFile(file);
						break;
                    }
                    catch(Throwable e)
					{
						break;
					}
                }
            }
        });

		btn_showNodeNames.setSelected(false);
		btn_showLinkIds.setSelected(false);
		btn_showNonConnectedNodes.setSelected(true);
	}

	@Override
	public void actionPerformed(ActionEvent e)
	{
		Object src = e.getSource();

		if (src == btn_load)
		{
			loadDesign();
		}
		else if (src == btn_loadDemand)
		{
			loadTrafficDemands();
		}
		else if (src == btn_save)
		{
			saveDesign();
		}
		else if (src == btn_showNodeNames)
		{
			canvas.showNodeNames(btn_showNodeNames.isSelected());
		}
		else if (src == btn_showLinkIds)
		{
			canvas.showLinkIds(btn_showLinkIds.isSelected());
		}
		else if (src == btn_showNonConnectedNodes)
		{
			canvas.showNonConnectedNodes(btn_showNonConnectedNodes.isSelected());
		}
		else if (src == btn_takeSnapshot)
		{
			takeSnapshot();
		}
		else if (src == btn_zoomIn)
		{
			zoomIn();
		}
		else if (src == btn_zoomOut)
		{
			zoomOut();
		}
		else if (src == btn_zoomAll)
		{
			zoomAll();
		}
		else if (src == btn_reset)
		{
			callback.reset();
		}
	}

	/**
	 * Adds a new plugin to the canvas.
	 *
	 * @param plugin Plugin to be added
	 * @since 0.3.0
	 */
	public void addPlugin(ITopologyCanvasPlugin plugin)
	{
		canvas.addPlugin(plugin);
	}
	
	private void checkNetPlanFileChooser()
	{
		if (fc_netPlan == null)
		{
			fc_netPlan = new FileChooserNetworkDesign(defaultDesignDirectory, DialogType.NETWORK_DESIGN);
		}
	}

	private void checkDemandFileChooser()
	{
		if (fc_demands == null)
		{
			fc_demands = new FileChooserNetworkDesign(defaultDemandDirectory, DialogType.DEMANDS);
		}
	}

	/**
	 * Returns a reference to the topology canvas.
	 * 
	 * @return Reference to the topology canvas
	 * @since 0.2.3
	 */
	public ITopologyCanvas getCanvas()
	{
		return canvas;
	}

	/**
	 * Loads a network design from a <code>.n2p</code> file.
	 *
	 * @since 0.3.0
	 */
	public void loadDesign()
	{
		try
		{
			checkNetPlanFileChooser();

			int rc = fc_netPlan.showOpenDialog(null);
			if (rc != JFileChooser.APPROVE_OPTION) return;

			NetPlan aux = fc_netPlan.readNetPlan();
			callback.loadDesign(aux);
		}
		catch (Net2PlanException ex)
		{
			if (ErrorHandling.isDebugEnabled()) ErrorHandling.addErrorOrException(ex, TopologyPanel.class);
			ErrorHandling.showErrorDialog(ex.getMessage(), "Error loading network design");
		}
		catch (Throwable ex)
		{
			ErrorHandling.addErrorOrException(ex, TopologyPanel.class);
			ErrorHandling.showErrorDialog("Error loading network design");
		}
	}
	
	private void loadDesignFromFile(File file)
	{
		try
		{
			NetPlan netPlan = new NetPlan(file);
			
			checkNetPlanFileChooser();
			fc_netPlan.setCurrentDirectory(file.getParentFile());			

			callback.loadDesign(netPlan);
		}
		catch (Net2PlanException ex)
		{
			if (ErrorHandling.isDebugEnabled()) ErrorHandling.addErrorOrException(ex, TopologyPanel.class);
			ErrorHandling.showErrorDialog(ex.getMessage(), "Error loading network design");
		}
		catch (Throwable ex)
		{
			ErrorHandling.addErrorOrException(ex, TopologyPanel.class);
			ErrorHandling.showErrorDialog("Error loading network design");
		}
	}

	/**
	 * Loads traffic demands from a <code>.n2p</code> file, overriding current demands.
	 *
	 * @since 0.3.0
	 */
	public void loadTrafficDemands()
	{
		try
		{
			checkDemandFileChooser();

			int rc = fc_demands.showOpenDialog(null);
			if (rc != JFileChooser.APPROVE_OPTION) return;

			NetPlan aux_demands = new NetPlan(fc_demands.getSelectedFile());
			callback.loadTrafficDemands(aux_demands);
		}
		catch (Net2PlanException ex)
		{
			if (ErrorHandling.isDebugEnabled()) ErrorHandling.addErrorOrException(ex, TopologyPanel.class);
			ErrorHandling.showErrorDialog(ex.getMessage(), "Error loading traffic demands");
		}
		catch (Exception ex)
		{
			ErrorHandling.addErrorOrException(ex, TopologyPanel.class);
			ErrorHandling.showErrorDialog("Error loading traffic demands");
		}
	}

	/**
	 * Saves a network design to a <code>.n2p</code> file.
	 *
	 * @since 0.3.0
	 */
	public void saveDesign()
	{
		try
		{
			checkNetPlanFileChooser();

			int rc = fc_netPlan.showSaveDialog(null);
			if (rc != JFileChooser.APPROVE_OPTION) return;

			NetPlan netPlan = callback.getDesign();
			if (netPlan.isEmpty()) throw new Net2PlanException("Design is empty");

			fc_netPlan.saveNetPlan(netPlan);
			ErrorHandling.showInformationDialog("Design saved successfully", "Save design");
		}
		catch (Net2PlanException ex)
		{
			if (ErrorHandling.isDebugEnabled()) ErrorHandling.addErrorOrException(ex, TopologyPanel.class);
			ErrorHandling.showErrorDialog(ex.getMessage(), "Error saving network design");
		}
		catch (Throwable ex)
		{
			ErrorHandling.addErrorOrException(ex, TopologyPanel.class);
			ErrorHandling.showErrorDialog("Error saving network design");
		}
	}
	
	/**
	 * Configures the topology panel to allow (or not) loading of external traffic demand files.
	 * 
	 * @param isAllowed Indicates whether or not it is allowed to load traffic demand files.
	 * @since 0.3.0
	 */
	public void setAllowLoadTrafficDemand(boolean isAllowed)
	{
		btn_loadDemand.setVisible(isAllowed);
	}

	/**
	 * Take a snapshot of the canvas.
	 * 
	 * @since 0.3.0
	 */
	public void takeSnapshot()
	{
		canvas.takeSnapshot();
	}
	
	/**
	 * Makes zoom-all from the center of the view.
	 * 
	 * @since 0.3.0
	 */
	public void zoomAll()
	{
		canvas.zoomAll();
	}
	
	/**
	 * Makes zoom-in from the center of the view.
	 * 
	 * @since 0.3.0
	 */
	public void zoomIn()
	{
		canvas.zoomIn();
	}

	/**
	 * Makes zoom-out from the center of the view.
	 * 
	 * @since 0.3.0
	 */
	public void zoomOut()
	{
		canvas.zoomOut();
	}
}