/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.internal.plugins;

import com.net2plan.gui.utils.FileChooserConfirmOverwrite;
import com.net2plan.gui.utils.topology.ITopologyCanvasPlugin;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.internal.ImageUtils;
import com.net2plan.utils.Pair;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileNameExtensionFilter;

/**
 * Base class for topology canvas.
 * 
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.3
 */
public abstract class ITopologyCanvas implements Plugin
{
	private JFileChooser fc;
	
	/**
	 * Adds a new unidirectional link between two nodes.
	 * 
	 * @param linkId Link identifier
	 * @param originNodeId Origin node identifier
	 * @param destinationNodeId Destinatin node identifier
	 * @since 0.3.0
	 */
	public abstract void addLink(long linkId, long originNodeId, long destinationNodeId);

	/**
	 * Adds a new node.
	 * 
	 * @param nodeId Node identifier
	 * @param pos Node position
	 * @param label Node label (it can be null)
	 * @since 0.3.0
	 */
	public abstract void addNode(long nodeId, Point2D pos, String label);

	/**
	 * Adds a new plugin to the canvas.
	 * 
	 * @param plugin Plugin
	 * @since 0.3.0
	 */
	public abstract void addPlugin(ITopologyCanvasPlugin plugin);

	/**
	 * Returns the real coordinates in the topology for a given screen point.
	 * 
	 * @param screenPoint Screen location
	 * @return Coordinates in the topology system for the screen point
	 * @since 0.3.0
	 */
	public abstract Point2D convertViewCoordinatesToRealCoordinates(Point screenPoint);
	
	/**
	 * Decrease the font size.
	 * 
	 * @since 0.3.0
	 */
	public abstract void decreaseFontSize();
	
	/**
	 * Decrease the node size.
	 * 
	 * @since 0.3.0
	 */
	public abstract void decreaseNodeSize();
	
	/**
	 * Returns the top-level component of the canvas.
	 * 
	 * @return Top-level component of the canvas
	 * @since 0.3.0
	 */
	public abstract JComponent getComponent();
	
	/**
	 * Returns a reference to the internal component containing the canvas.
	 *
	 * @return Internal component containing the canvas
	 * @since 0.3.0
	 */
	public abstract JComponent getInternalComponent();
	
	/**
	 * Increase the font size.
	 * 
	 * @since 0.3.0
	 */
	public abstract void increaseFontSize();
	
	/**
	 * Increase the node size.
	 * 
	 * @since 0.3.0
	 */
	public abstract void increaseNodeSize();
	
	/**
	 * Indicates whether a link is visible or not.
	 * 
	 * @param linkId Link identifier
	 * @return {@code true} if the link is visible. Otherwise, {@code false}
	 * @since 0.3.0
	 */
	public abstract boolean isLinkVisible(long linkId);

	/**
	 * Indicates whether a node is visible or not.
	 * 
	 * @param nodeId Node identifier
	 * @return {@code true} if the node is visible. Otherwise, {@code false}
	 * @since 0.3.0
	 */
	public abstract boolean isNodeVisible(long nodeId);

	/**
	 * Refreshes the canvas.
	 * 
	 * @since 0.3.0
	 */
	public abstract void refresh();

	/**
	 * Removes a link from the canvas.
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public abstract void removeLink(long linkId);

	/**
	 * Removes a node from the canvas, and all its associated incoming/outgoing links.
	 * 
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public abstract void removeNode(long nodeId);

	/**
	 * Removes a plugin from the canvas.
	 * 
	 * @param plugin Plugin
	 * @since 0.3.0
	 */
	public abstract void removePlugin(ITopologyCanvasPlugin plugin);

	/**
	 * Resets the graph.
	 * 
	 * @since 0.3.0
	 */
	public abstract void reset();

	/**
	 * Resets the elements down.
	 * 
	 * @since 0.3.0
	 */
	public abstract void resetDownState();

	/**
	 * Resets the emphasized elements.
	 * 
	 * @since 0.3.0
	 */
	public abstract void resetPickedState();

	/**
	 * Sets all links are visible/hidden.
	 * 
	 * @param visible Indicates whether all links are visible (<code>true</code>) or hidden (<code>false</code>)
	 * @since 0.3.0
	 */
	public abstract void setAllLinksVisible(boolean visible);
	
	/**
	 * Sets all nodes are visible/hidden.
	 * 
	 * @param visible Indicates whether all nodes are visible (<code>true</code>) or hidden (<code>false</code>)
	 * @since 0.3.0
	 */
	public abstract void setAllNodesVisible(boolean visible);
	
	/**
	 * Sets a link to down state.
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public abstract void setLinkDown(long linkId);

	/**
	 * Sets a link to up state.
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public abstract void setLinkUp(long linkId);

	/**
	 * Sets a link as visible/hidden.
	 * 
	 * @param linkId Link identifier
	 * @param visible Indicates whether the link is visible (<code>true</code>) or hidden (<code>false</code>)
	 * @since 0.3.0
	 */
	public abstract void setLinkVisible(long linkId, boolean visible);

	/**
	 * Sets some links as visible/hidden.
	 * 
	 * @param linkIds Link identifiers
	 * @param visible Indicates whether the links are visible (<code>true</code>) or hidden (<code>false</code>)
	 * @since 0.3.0
	 */
	public abstract void setLinksVisible(Collection<Long> linkIds, boolean visible);
	
	/**
	 * Sets a node to down state.
	 * 
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public abstract void setNodeDown(long nodeId);
	
	/**
	 * Sets the position of a node.
	 * 
	 * @param nodeId Node identifier
	 * @param position Node position
	 * @since 0.3.0
	 */
	public abstract void setNodeXYPosition(long nodeId, Point2D position);

	/**
	 * Sets a node to up state.
	 *
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public abstract void setNodeUp(long nodeId);

	/**
	 * Sets a node as visible/hidden.
	 * 
	 * @param nodeId Node identifier
	 * @param visible Indicates whether the node is visible (<code>true</code>) or hidden (<code>false</code>)
	 * @since 0.3.0
	 */
	public abstract void setNodeVisible(long nodeId, boolean visible);
	
	/**
	 * Sets some nodes as visible/hidden.
	 * 
	 * @param nodeIds Node identifiers
	 * @param visible Indicates whether the nodes are visible (<code>true</code>) or hidden (<code>false</code>)
	 * @since 0.3.0
	 */
	public abstract void setNodesVisible(Collection<Long> nodeIds, boolean visible);
	
	/**
	 * Indicates whether or not non-connected nodes should be shown.
	 * 
	 * @param show Indicates whether or not non-connected nodes should be shown
	 * @since 0.3.0
	 */
	public abstract void showNonConnectedNodes(boolean show);

	/**
	 * Emphasizes a link.
	 * 
	 * @param linkId Link identifier
	 * @since 0.3.0
	 */
	public abstract void showLink(long linkId);

	/**
	 * Indicates whether or not link identifiers should be shown.
	 * 
	 * @param show Indicates whether or not link identifiers should be shown
	 * @since 0.3.0
	 */
	public abstract void showLinkIds(boolean show);
	
	/**
	 * Emphasizes a node.
	 * 
	 * @param nodeId Node identifier
	 * @since 0.3.0
	 */
	public abstract void showNode(long nodeId);

	/**
	 * Indicates whether or not node names should be shown.
	 * 
	 * @param show Indicates whether or not node names should be shown
	 * @since 0.3.0
	 */
	public abstract void showNodeNames(boolean show);

	/**
	 * Emphasizes a set of nodes.
	 * 
	 * @param nodeIds Node identifiers
	 * @since 0.3.0
	 */
	public abstract void showNodes(Collection<Long> nodeIds);

	/**
	 * Emphasizes a set of nodes and/or links.
	 * 
	 * @param nodeIds Node identifiers (may be null)
	 * @param linkIds Link identifiers (may be null)
	 * @since 0.3.0
	 */
	public abstract void showNodesAndLinks(Collection<Long> nodeIds, Collection<Long> linkIds);

	/**
	 * Emphasizes a set of links.
	 * 
	 * @param linkIds Links identifiers
	 * @since 0.3.0
	 */
	public abstract void showRoute(List<Long> linkIds);

	/**
	 * Emphasizes two sets of links. The first one will show a heavy line, while
	 * the second one will show a heavy dashed line.
	 * 
	 * @param primaryRouteLinks Link identifiers for the first set of links
	 * @param secondaryRouteLinks Link identifiers for the second set of links
	 * @since 0.3.0
	 */
	public abstract void showRoutes(Collection<Long> primaryRouteLinks, Collection<Long> secondaryRouteLinks);

	/**
	 * Takes a snapshot of the canvas.
	 * 
	 * @since 0.3.0
	 */
	public final void takeSnapshot()
	{
		if (fc == null)
		{
			fc = new FileChooserConfirmOverwrite();
			FileNameExtensionFilter pngFilter = new FileNameExtensionFilter("PNG files", "png");
			fc.setFileFilter(pngFilter);
		}

		takeSnapshot_preConfigure();
		JComponent component = getInternalComponent();
		BufferedImage bi = ImageUtils.takeSnapshot(component);
		takeSnapshot_postConfigure();
		
		int s = fc.showSaveDialog(null);
		if (s == JFileChooser.APPROVE_OPTION)
		{
			File f = fc.getSelectedFile();
			ImageUtils.writeImageToFile(f, bi, ImageUtils.ImageType.PNG);
		}
	}
	
	/**
	 * Configures the canvas before taking an snapshot (i.e. hiding elements).
	 * 
	 * @since 0.3.0
	 */
	public void takeSnapshot_preConfigure()
	{
	}

	/**
	 * Restores the canvas to its original state after taking an snapshot.
	 * 
	 * @since 0.3.0
	 */
	public void takeSnapshot_postConfigure()
	{
	}
	
	/**
	 * Refresh the canvas with the physical topology from the given network design.
	 * 
	 * @param netPlan Network design
	 * @since 0.3.0
	 */
	public final void updateTopology(NetPlan netPlan)
	{
		long layerId = netPlan.getLayerDefaultId();
		updateTopology(netPlan, layerId);
	}

	/**
	 * Refresh the canvas with the physical topology from the given network 
	 * design in the given layer.
	 * 
	 * @param netPlan Network design
	 * @param layerId Layer identifier
	 * @since 0.3.0
	 */
	public final void updateTopology(NetPlan netPlan, long layerId)
	{
		Map<Long, Point2D> nodeXYPositionMap = netPlan.getNodeXYPositionMap();
		Map<Long, String> nodeNameMap = netPlan.getNodeNameMap();
		Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap(layerId);
		
		updateTopology(nodeXYPositionMap, nodeNameMap, linkMap);
		
		Set<Long> nodesDown = netPlan.getNodesDown();
		Set<Long> linksDown = netPlan.getLinksDown(layerId);
		for(long nodeId : nodesDown) setNodeDown(nodeId);
		for(long linkId : linksDown) setLinkDown(linkId);
	}
	
	/**
	 * Refresh the canvas with the given physical topology.
	 * 
	 * @param nodeXYPositionMap Map of node XY position
	 * @param nodeNameMap Map of node names
	 * @param linkMap Map of links
	 * @since 0.3.0
	 */
	public abstract void updateTopology(Map<Long, Point2D> nodeXYPositionMap, Map<Long, String> nodeNameMap, Map<Long, Pair<Long, Long>> linkMap);

	/**
	 * Makes zoom-all from the center of the view.
	 * 
	 * @since 0.3.0
	 */
	public abstract void zoomAll();

	/**
	 * Makes zoom-in from the center of the view.
	 * 
	 * @since 0.3.0
	 */
	public abstract void zoomIn();

	/**
	 * Makes zoom-out from the center of the view.
	 * 
	 * @since 0.3.0
	 */
	public abstract void zoomOut();
}
