/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.internal.sim;

import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.interfaces.simulation.ICombinedEventGenerator;
import com.net2plan.interfaces.simulation.ICombinedEventProcessor;
import com.net2plan.interfaces.simulation.IEventGenerator;
import com.net2plan.interfaces.simulation.IEventProcessor;
import com.net2plan.interfaces.simulation.SimEvent;
import com.net2plan.internal.IExternal;
import com.net2plan.utils.StringUtils;
import java.util.Map;
import javax.swing.JComponent;

/**
 * Simulates the network operation, where failures in links and nodes randomly
 * appear according to the user-defined reliability information and the
 * definition of shared risk groups (SRG). Targeted to evaluate availability
 * performances of built-in or user-defined protection/restoration schemes.
 * 
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.2
 */
public class OnlineSimulation extends SimKernel
{
	private NetPlan currentNetPlan, currentNetPlan_view;
	private SimStats stats;

	@Override
	public void checkLoadedNetPlan(NetPlan netPlan)
	{
	}

	@Override
	public void finishTransitory(double currentSimTime)
	{
		((IEventGenerator) eventGenerator).finishTransitory(currentSimTime);
		((IEventProcessor) eventProcessor).finishTransitory(currentSimTime);

		if (!disableStatistics) stats.reset(currentSimTime);
	}

	@Override
	public String getCommandLineSpecificHelp()
	{
		return "Simulates the network operation, where events appeared according to "
				+ "a built-in or user-defined event generator. Targeted to evaluate the performance of "
				+ "built-in or user-defined reaction algorithms";
	}

	@Override
	public NetPlan getCurrentNetPlan()
	{
		if (currentNetPlan_view == null) currentNetPlan_view = currentNetPlan.unmodifiableView();
		return currentNetPlan_view;
	}

	@Override
	public Class<? extends IExternal> getEventGeneratorClass()
	{
		return IEventGenerator.class;
	}

	@Override
	public String getEventGeneratorLabel()
	{
		return "Event generator";
	}

	@Override
	public Class<? extends IExternal> getEventProcessorClass()
	{
		return IEventProcessor.class;
	}

	@Override
	public String getEventProcessorLabel()
	{
		return "Provisioning algorithm";
	}

	@Override
	public String getModeName()
	{
		return "online-sim";
	}

	@Override
	public JComponent getPluginConfigurationScreen()
	{
		return null;
	}

	@Override
	public Map<String, String> getPluginDefaultOptions()
	{
		return null;
	}

	@Override
	public String getPluginName()
	{
		return "Online simulation (CLI)";
	}

	@Override
	public String getSimulationReport()
	{
		double simTime = getSimCore().getFutureEventList().getCurrentSimulationTime();
		double cpuTime = getSimCore().getCPUTime();
		long processedEvents = getSimCore().getFutureEventList().getNumberOfProcessedEvents();
		int pendingEvents = getSimCore().getFutureEventList().getNumberOfPendingEvents();
		double evToSecRatio = cpuTime == 0 ? 0 : (double) processedEvents / cpuTime;
		double simToWallRatio = cpuTime == 0 ? 0 : (double) simTime / cpuTime;
		double evToSimSecRatio = simTime == 0 ? 0 : (double) processedEvents / simTime;
		
		StringBuilder info = new StringBuilder();
		info.append("<html><head><title>Simulation report</title></head>");
		info.append("<body>");
		info.append("<h1>Simulation information</h1>");
		info.append("<center><table border='1'><tr><th>Parameter</th><th>Value</th></tr>");
		info.append(String.format("<tr><td>Current simulation time</td><td>%s</td></tr>", StringUtils.secondsToYearsDaysHoursMinutesSeconds(simTime)));
		info.append(String.format("<tr><td>Current CPU time</td><td>%s (%.3g simsec/sec)</td></tr>", StringUtils.secondsToYearsDaysHoursMinutesSeconds(cpuTime), simToWallRatio));
		info.append(String.format("<tr><td>Number of processed events</td><td>%d (%.3g ev/sec, %.3g ev/simsec)</td></tr>", processedEvents, evToSecRatio, evToSimSecRatio));
		info.append(String.format("<tr><td>Number of pending events</td><td>%d</td></tr>", pendingEvents));
		info.append("</table></center>");

		info.append("<h1>General results</h1>");

		if (stats == null)
		{
			info.append("<p>No results available since 'disableStatistics' was set to 'true'</p>");
		}
		else
		{
			info.append(stats.getResults(getSimCore().getFutureEventList().getCurrentSimulationTime()));
		}

		info.append("<h1>Problem-specific results</h1>");
		
		StringBuilder eventGeneratorHtml = new StringBuilder();
		String eventGeneratorOut = ((IEventGenerator) eventGenerator).finish(eventGeneratorHtml, simTime);

		if (eventGeneratorOut != null && eventGeneratorHtml.length() > 0)
		{
			info.append(String.format("<h2>%s</h2>", eventGeneratorOut));
			info.append(eventGeneratorHtml);
		}
		else
		{
			info.append("<p>No results from the event generator");
		}
		
		StringBuilder eventProcessorHtml = new StringBuilder();
		String eventProcessorOut = ((IEventProcessor) eventProcessor).finish(eventProcessorHtml, simTime);

		if (eventProcessorOut != null && eventProcessorHtml.length() > 0)
		{
			info.append(String.format("<h2>%s</h2>", eventProcessorOut));
			info.append(eventProcessorHtml);
		}
		else
		{
			info.append("<p>No results from the event processor");
		}

		info.append("</body></html>");

		return info.toString();
	}

	@Override
	public void initialize()
	{
		((ISimExternal) eventGenerator).setKernel(this);
		if (eventGenerator instanceof ICombinedEventGenerator)
			for(IEventGenerator internalEventGenerator : ((ICombinedEventGenerator) eventGenerator).eventGenerators)
				((ISimExternal) internalEventGenerator).setKernel(this);
		
		((ISimExternal) eventProcessor).setKernel(this);
		if (eventProcessor instanceof ICombinedEventProcessor)
			for(IEventProcessor internalEventProcessor : ((ICombinedEventProcessor) eventProcessor).eventProcessors)
				((ISimExternal) internalEventProcessor).setKernel(this);
		
		((ISimExternal) eventGenerator).initialize(initialNetPlan, currentNetPlan, eventGeneratorParameters, simulationParameters, net2planParameters);
		((ISimExternal) eventProcessor).initialize(initialNetPlan, currentNetPlan, eventProcessorParameters, simulationParameters, net2planParameters);
		if (!disableStatistics) stats = new SimStats(initialNetPlan, currentNetPlan.unmodifiableView(), simulationParameters, net2planParameters);
	}
	
	@Override
	public void initializeNetState()
	{
		if (getSimCore().getSimulationState() != SimCore.SimState.NOT_STARTED)
			throw new Net2PlanException("Network state cannot be re-initialized once the simulation was started");

		currentNetPlan = initialNetPlan.copy();
		currentNetPlan_view = null;
	}

	@Override
	public void simulationLoop(SimEvent event)
	{
		currentNetPlan_view = null;
		
		switch(event.getEventDestinationModule())
		{
			case EVENT_GENERATOR:
				((ISimExternal) eventGenerator).processEvent(currentNetPlan, event);
				break;
				
			case EVENT_PROCESSOR:
				((ISimExternal) eventProcessor).processEvent(currentNetPlan, event);
				break;
				
			default:
				throw new RuntimeException("Bad");
		}

		if (!disableStatistics) stats.computeNextState(event.getEventTime());
	}

	@Override
	public void resetModule()
	{
		stats = null;
	}

}
