/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.internal.sim;

import com.net2plan.interfaces.simulation.SimEvent;
import com.jom.JOMException;
import com.net2plan.interfaces.networkDesign.Configuration;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.internal.CommandLineParser;
import com.net2plan.internal.CustomHTMLEditorKit;
import com.net2plan.internal.ErrorHandling;
import com.net2plan.internal.plugins.ICLIModule;
import com.net2plan.internal.IExternal;
import com.net2plan.utils.ClassLoaderUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.StringUtils;
import com.net2plan.utils.Triple;
import java.io.File;
import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PatternOptionBuilder;

/**
 * Core-class for simulators. Users are only responsible to implement their
 * simulation-specific methods (check input data, process events...).
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @since 0.2.2
 */
public abstract class SimKernel implements IEventCallback, ICLIModule
{
	private final static DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");

	/**
	 * Reference to the initial network design.
	 * 
	 * @since 0.2.2
	 */
	protected NetPlan initialNetPlan;

	/**
	 * A key-value map with simulation parameters.
	 * 
	 * @since 0.3.0
	 */
	protected Map<String, String> simulationParameters;

	/**
	 * A key-value map with {@code Net2Plan}-wide configuration options.
	 * 
	 * @since 0.2.2
	 */
	protected Map<String, String> net2planParameters;

	/**
	 * A key-value map with specific parameters for the event generator.
	 * 
	 * @since 0.2.2
	 */
	protected Map<String, String> eventGeneratorParameters;

	/**
	 * A key-value map with specific parameters for the event processor.
	 * 
	 * @since 0.2.2
	 */
	protected Map<String, String> eventProcessorParameters;
	
	/**
	 * Reference to the event generator.
	 * 
	 * @since 0.2.2
	 */
	protected IExternal eventGenerator;
	
	/**
	 * Reference to the event processor.
	 * 
	 * @since 0.2.2
	 */
	protected IExternal eventProcessor;
	
	/**
	 * Indicates whether or not general statistics are enabled, or only specific 
	 * algorithm metrics are collected.
	 *
	 * @since 0.2.2
	 */
	protected boolean disableStatistics;

	private SimEvent lastEvent;
	private boolean omitProtectionSegments;
	private IGUISimulationListener guiListener;
	private Throwable lastReason = null;
	private final SimCore simCore;
	private NetPlan originalNetPlan;
	
	/**
	 * Default constructor.
	 * 
	 * @since 0.2.2
	 */
	public SimKernel()
	{
		simCore = new SimCore(this);
		setNetPlan(new NetPlan());
	}

	@Override
	public final void executeFromCommandLine(String[] args) throws ParseException
	{
		Options options = getCommandLineOptions();

		Class<? extends IExternal> eventGeneratorClass = getEventGeneratorClass();
		Class<? extends IExternal> eventProcessorClass = getEventProcessorClass();

		CommandLineParser parser = new CommandLineParser();
		CommandLine cli = parser.parse(options, args);

		/* 1. Read options file */
		if (cli.hasOption("config-file"))
		{
			try { Configuration.readFromOptionsFile((File) cli.getParsedOptionValue("config-file")); }
			catch (IOException e) {	throw new ParseException("Options file not loaded"); }
		}

		net2planParameters = Configuration.getOptions();

		/* 2. Load event generator and event processor objects */
		File generatorClassFile = (File) cli.getParsedOptionValue("generator-class-file");
		String generatorClassName = (String) cli.getParsedOptionValue("generator-class-name");
		File provisioningClassFile = (File) cli.getParsedOptionValue("processor-class-file");
		String provisioningClassName = (String) cli.getParsedOptionValue("processor-class-name");

		IExternal aux_eventGenerator = ClassLoaderUtils.getInstance(generatorClassFile, generatorClassName, eventGeneratorClass);
		IExternal aux_eventProcessor = ClassLoaderUtils.getInstance(provisioningClassFile, provisioningClassName, eventProcessorClass);

		/* 3. Read simulation, event generator and event processor parameters */
		Properties customSimulationParameters = cli.getOptionProperties("sim-param");
		Properties customEventGeneratorParameters = cli.getOptionProperties("generator-param");
		Properties customEventProcessorParameters = cli.getOptionProperties("processor-param");

		/* 4. Read the input netPlan file */
		File inputFile = (File) cli.getParsedOptionValue("input-file");
		File outputFile = (File) cli.getParsedOptionValue("output-file");
		
		/* 5. Initialize and run simulation */
		NetPlan aux_netPlan = new NetPlan(inputFile);
		String html = runSimulation(this, aux_netPlan, aux_eventGenerator, customEventGeneratorParameters, aux_eventProcessor, customEventProcessorParameters, customSimulationParameters, net2planParameters).getSecond();
		CustomHTMLEditorKit.exportToHTML(outputFile, html);
	}
	
	@Override
	public void endTransitory()
	{
		if (lastEvent == null) getSimCore().finishTransitory(0);
		else getSimCore().finishTransitory(lastEvent.getEventTime());
	}
	
	@Override
	public final String getCommandLineHelp()
	{
		StringBuilder out = new StringBuilder();
		out.append(getCommandLineSpecificHelp());
		out.append(StringUtils.getLineSeparator());
		out.append(StringUtils.getLineSeparator());
		out.append("This mode is built on top of a discrete-event simulator");
		return out.toString();
	}

	@Override
	public final Options getCommandLineOptions()
	{
		String generatorLabel = getEventGeneratorLabel();
		String processorLabel = getEventProcessorLabel();

		Options options = new Options();
		
		Option configFile = new Option(null, "config-file", true, "(Optional) configuration file with Net2Plan-wide parameters (by default, options.ini)");
		configFile.setType(PatternOptionBuilder.FILE_VALUE);
		configFile.setArgName("file");
		options.addOption(configFile);

		Option inputFile = new Option(null, "input-file", true, ".n2p file containing an initial network design");
		inputFile.setType(PatternOptionBuilder.FILE_VALUE);
		inputFile.setArgName("file");
		options.addOption(inputFile);

		Option outputFile = new Option(null, "output-file", true, "Output HTML file with the simulation report");
		outputFile.setType(PatternOptionBuilder.FILE_VALUE);
		outputFile.setArgName("file");
		outputFile.setRequired(true);
		options.addOption(outputFile);

		Option simParameters = new Option(null, "sim-param", true, "Simulation parameters (use one of this for each parameter)");
		simParameters.setArgName("property=value");
		simParameters.setArgs(2);
		simParameters.setValueSeparator('=');
		options.addOption(simParameters);

		Option generatorParameters = new Option(null, "generator-param", true, "(Optional) " + generatorLabel.toLowerCase(Locale.getDefault()) + " parameters (use one of this for each parameter)");
		generatorParameters.setArgName("property=value");
		generatorParameters.setArgs(2);
		generatorParameters.setValueSeparator('=');
		options.addOption(generatorParameters);
		
		Option processorParameters = new Option(null, "processor-param", true, "(Optional) " + processorLabel.toLowerCase(Locale.getDefault()) + " parameters (use one of this for each parameter)");
		processorParameters.setArgName("property=value");
		processorParameters.setArgs(2);
		processorParameters.setValueSeparator('=');
		options.addOption(processorParameters);

		Option eventGeneratorClassFile = new Option(null, "generator-class-file", true, ".class/.jar file containing the event generator");
		eventGeneratorClassFile.setType(PatternOptionBuilder.FILE_VALUE);
		eventGeneratorClassFile.setArgName("file");
		eventGeneratorClassFile.setRequired(true);
		options.addOption(eventGeneratorClassFile);

		Option eventGeneratorClassName = new Option(null, "generator-class-name", true, "Class name of the event generator (package name could be omitted)");
		eventGeneratorClassName.setType(PatternOptionBuilder.STRING_VALUE);
		eventGeneratorClassName.setArgName("classname");
		eventGeneratorClassName.setRequired(true);
		options.addOption(eventGeneratorClassName);

		Option eventProcessorClassFile = new Option(null, "processor-class-file", true, ".class/.jar file containing the event processor");
		eventProcessorClassFile.setType(PatternOptionBuilder.FILE_VALUE);
		eventProcessorClassFile.setArgName("file");
		eventProcessorClassFile.setRequired(true);
		options.addOption(eventProcessorClassFile);

		Option eventProcessorClassName = new Option(null, "processor-class-name", true, "Class name of the event processor (package name could be omitted)");
		eventProcessorClassName.setType(PatternOptionBuilder.STRING_VALUE);
		eventProcessorClassName.setArgName("classname");
		eventProcessorClassName.setRequired(true);
		options.addOption(eventProcessorClassName);

		return options;
	}
	
	@Override
	public final void processEvent(SimEvent event)
	{
		lastEvent = event;
		simulationLoop(event);
	}

	@Override
	public void refresh(boolean forceRefresh)
	{
		if (guiListener != null)
		{
			guiListener.refresh(forceRefresh);
		}
		else
		{
			System.out.println(getSimulationInfo());
		}
	}

	@Override
	public void scheduleEvent(SimEvent event)
	{
		simCore.getFutureEventList().addEvent(event);
	}

	@Override
	public void simulationStateChanged(SimCore.SimState simulationState, Throwable reason)
	{
		this.lastReason = reason;

		if (guiListener != null)
		{
			guiListener.simulationStateChanged(simulationState, reason);
		}
		else
		{
			if (reason == null || reason instanceof EndSimulationException) return;

			if (reason instanceof Net2PlanException || reason instanceof JOMException)
			{
				if (ErrorHandling.isDebugEnabled()) ErrorHandling.addErrorOrException(reason, SimKernel.class);
				
				System.out.println("Error executing simulation");
				System.out.println();
				System.out.println(reason.getMessage());
			}
			else
			{
				System.out.println("Fatal error");
				System.out.println();
				ErrorHandling.printStackTrace(reason);
			}
		}
	}
	
	/**
	 * Checks for validity of the loaded network design.
	 *
	 * @param netPlan An input network design
	 * @since 0.2.2
	 */
	public abstract void checkLoadedNetPlan(NetPlan netPlan);

	/**
	 * Configures the simulation.
	 * 
	 * @param simulationParameters Simulation parameters
	 * @param net2planParameters Net2Plan-wide configuration parameters
	 * @param eventGenerator An instance of the event generator
	 * @param eventGeneratorParameters Parameter-value map for the event generator
	 * @param eventProcessor An instance of the event processor
	 * @param eventProcessorParameters Parameter-value map for the event processor
	 * @since 0.2.2
	 */
	public void configureSimulation(Map<String, String> simulationParameters, Map<String, String> net2planParameters, IExternal eventGenerator, Map<String, String> eventGeneratorParameters, IExternal eventProcessor, Map<String, String> eventProcessorParameters)
	{
		this.simulationParameters = new LinkedHashMap<String, String>(simulationParameters);

		if (!simulationParameters.containsKey("disableStatistics"))	throw new Net2PlanException("'disableStatistics' parameter is not configured");
		disableStatistics = Boolean.parseBoolean(simulationParameters.get("disableStatistics"));

		if (!simulationParameters.containsKey("omitProtectionSegments")) throw new Net2PlanException("'omitProtectionSegments' parameter is not configured");
		omitProtectionSegments = Boolean.parseBoolean(simulationParameters.get("omitProtectionSegments"));

		originalNetPlan = initialNetPlan;

		if (omitProtectionSegments && initialNetPlan.hasProtectionSegments())
		{
			NetPlan aux_netPlan = initialNetPlan.copy();
			aux_netPlan.removeAllProtectionSegments();
			setNetPlan(aux_netPlan);
		}

		if (!simulationParameters.containsKey("refreshTime")) throw new Net2PlanException("'refreshTime' parameter is not configured");
		double refreshTimeInSeconds = Double.parseDouble(simulationParameters.get("refreshTime"));
		simCore.setRefreshTimeInSeconds(refreshTimeInSeconds);

		if (!simulationParameters.containsKey("simEvents")) throw new Net2PlanException("'simEvents' parameter is not configured");
		long simEvents = Long.parseLong(simulationParameters.get("simEvents"));
		simCore.setTotalSimulationEvents(simEvents);

		if (!simulationParameters.containsKey("transitoryEvents")) throw new Net2PlanException("'transitoryEvents' parameter is not configured");
		long transitoryEvents = Long.parseLong(simulationParameters.get("transitoryEvents"));
		simCore.setTotalTransitoryEvents(transitoryEvents);

		if (!simulationParameters.containsKey("transitoryTime")) throw new Net2PlanException("'transitoryTime' parameter is not configured");
		double transitoryTime = Double.parseDouble(simulationParameters.get("transitoryTime"));
		simCore.setTotalTransitoryTime(transitoryTime);

		if (!simulationParameters.containsKey("simTime")) throw new Net2PlanException("'simTime' parameter is not configured");
		double simTime = Double.parseDouble(simulationParameters.get("simTime"));
		simCore.setTotalSimulationTime(simTime);
		
		if (!getEventGeneratorClass().isAssignableFrom(eventGenerator.getClass())) throw new RuntimeException("Bad - Event generator is not an instance of " + getEventGeneratorClass().getName());
		if (!getEventProcessorClass().isAssignableFrom(eventProcessor.getClass())) throw new RuntimeException("Bad - Event processor is not an instance of " + getEventProcessorClass().getName());

		this.net2planParameters = new LinkedHashMap<String, String>(net2planParameters);
		this.eventGenerator = eventGenerator;
		this.eventGeneratorParameters = new LinkedHashMap<String, String>(eventGeneratorParameters);
		this.eventProcessor = eventProcessor;
		this.eventProcessorParameters = new LinkedHashMap<String, String>(eventProcessorParameters);
	}

	/**
	 * Returns the command 
	 * @return
	 * @since 0.2.2
	 */
	public abstract String getCommandLineSpecificHelp();

	/**
	 * Returns the current network plan corresponding to the network state.
	 * 
	 * @return Current network plan corresponding to the network state
	 * @since 0.2.2
	 */
	public abstract NetPlan getCurrentNetPlan();

	/**
	 * Returns the class of the event generator.
	 * 
	 * @return Class of the event generator
	 * @since 0.2.2
	 */
	public abstract Class<? extends IExternal> getEventGeneratorClass();

	/**
	 * Returns the label for the event generator.
	 * 
	 * @return Label for the event generator
	 * @since 0.2.2
	 */
	public abstract String getEventGeneratorLabel();

	/**
	 * Returns the class of the event processor.
	 * 
	 * @return Class of the event processor
	 * @since 0.2.2
	 */
	public abstract Class<? extends IExternal> getEventProcessorClass();

	/**
	 * Returns the label for the event processor.
	 * 
	 * @return Label for the event processor
	 * @since 0.2.2
	 */
	public abstract String getEventProcessorLabel();
	
	/**
	 * Returns the initial network plan.
	 * 
	 * @return The initial network plan
	 * @since 0.2.2
	 */
	public NetPlan getInitialNetPlan()
	{
		return initialNetPlan;
	}

	@Override
	public abstract String getModeName();

	/**
	 * Returns a reference to the simulation core.
	 * 
	 * @return Reference to the simulation core
	 * @since 0.2.0
	 */
	public SimCore getSimCore()
	{
		return simCore;
	}

	/**
	 * Returns a brief simulation information report (current simulation time,
	 * last event processed...).
	 * 
	 * @return Simulation information
	 * @since 0.2.2
	 */
	public String getSimulationInfo()
	{
		if (simCore.getSimulationState() == SimCore.SimState.NOT_STARTED)
		{
			return "Simulation not started yet";
		}

		String NEWLINE = StringUtils.getLineSeparator();

		double simTime = simCore.getFutureEventList().getCurrentSimulationTime();
		double cpuTime = simCore.getCPUTime();
		long processedEvents = simCore.getFutureEventList().getNumberOfProcessedEvents();
		int pendingEvents = simCore.getFutureEventList().getNumberOfPendingEvents();
		double evToSecRatio = cpuTime == 0 ? 0 : (double) processedEvents / cpuTime;
		double simToWallRatio = cpuTime == 0 ? 0 : (double) simTime / cpuTime;
		double evToSimSecRatio = simTime == 0 ? 0 : (double) processedEvents / simTime;
		
		StringBuilder info = new StringBuilder();
		info.append(String.format("Current date: %s", dateFormat.format(Calendar.getInstance().getTime())));
		info.append(NEWLINE);
		info.append(String.format("Current simulation time: %s", StringUtils.secondsToYearsDaysHoursMinutesSeconds(simTime)));
		info.append(NEWLINE);
		info.append(String.format("Current CPU time: %s (%.3g simsec/sec)", StringUtils.secondsToYearsDaysHoursMinutesSeconds(cpuTime), simToWallRatio));
		info.append(NEWLINE);
		info.append(String.format("Number of processed events: %d (%.3g ev/sec, %.3g ev/simsec)", processedEvents, evToSecRatio, evToSimSecRatio));
		info.append(NEWLINE);
		info.append(String.format("Number of pending events: %d", pendingEvents));
		info.append(NEWLINE);
		info.append(String.format("Last event processed: %s", lastEvent == null ? "None" : lastEvent.toString()));
		info.append(NEWLINE);

		return info.toString();
	}

	/**
	 * Returns the parameters for the simulation.
	 * 
	 * @return Parameters for the simulation
	 * @since 0.2.2
	 */
	public final List<Triple<String, String, String>> getSimulationParameters()
	{
		List<Triple<String, String, String>> parameters = new LinkedList<Triple<String, String, String>>();
		parameters.add(Triple.of("disableStatistics", "#boolean# false", "Disable compilation of simulation statistics (only simulation information, and optionally algorithm-specific information, is collected)"));
		parameters.add(Triple.of("refreshTime", "10", "Refresh time (in seconds)"));
		parameters.add(Triple.of("omitProtectionSegments", "#boolean# false", "Remove protection segments from the network plan to free their reserved bandwidth"));
		parameters.add(Triple.of("simEvents", "-1", "Total simulation events (including transitory period) (-1 means no limit). In case that 'simTime' and 'simEvents' are specified, the transitory period will finish when one of the previous values is reached"));
		parameters.add(Triple.of("transitoryEvents", "-1", "Number of events for transitory period (-1 means no transitory period). In case that 'transitoryTime' and 'transitoryEvents' are specified, the transitory period will finish when one of the previous values is reached"));
		parameters.add(Triple.of("simTime", "-1", "Total simulation time (in seconds, including transitory period) (-1 means no limit). In case that 'simTime' and 'simEvents' are specified, the transitory period will finish when one of the previous values is reached"));
		parameters.add(Triple.of("transitoryTime", "-1", "Transitory time (in seconds) (-1 means no transitory period). In case that 'transitoryTime' and 'transitoryEvents' are specified, the transitory period will finish when one of the previous values is reached"));

		List<Triple<String, String, String>> simSpecificParameters = getSimulationSpecificParameters();
		if (simSpecificParameters != null)
		{
			simSpecificParameters = new LinkedList<Triple<String, String, String>>(simSpecificParameters);
			Iterator<Triple<String, String, String>> it = simSpecificParameters.iterator();
			while(it.hasNext())
			{
				String newParamName = it.next().getFirst();
				for(Triple<String, String, String> parameter : parameters)
				{
					if (parameter.getFirst().equalsIgnoreCase(newParamName))
					{
						it.remove();
						break;
					}
				}
			}
		}

		return parameters;
	}

	/**
	 * Returns the simulation report.
	 * 
	 * @return Simulation report
	 * @since 0.2.2
	 */
	public abstract String getSimulationReport();
	
	/**
	 * Returns the parameters for a specific simulation tool. It should not be
	 * called externally, since it is already called by {@link #getSimulationParameters() getSimulationParameters} method.
	 * 
	 * @return Parameters for a specific simulation tool
	 * @since 0.2.2
	 */
	protected List<Triple<String, String, String>> getSimulationSpecificParameters()
	{
		return new LinkedList<Triple<String, String, String>>();
	}

	/**
	 * Initialize the simulation (event generator, event processor...). In case 
	 * that new events are to be generated, they should be scheduled calling to 
	 * {@link #scheduleEvent(com.net2plan.interfaces.simulation.SimEvent) scheduleEvent}.
	 * 
	 * @since 0.3.0
	 */
	public abstract void initialize();

	/**
	 * Initializes the current network state from a initial network plan.
	 * 
	 * @since 0.2.2
	 */
	public abstract void initializeNetState();

	/**
	 * Resets the simulation.
	 * 
	 * @since 0.2.2
	 */
	public void reset()
	{
		simCore.reset();

		if (originalNetPlan != null)
		{
			setNetPlan(originalNetPlan);
			originalNetPlan = null;
		}
		else
		{
			initializeNetState();
		}

		lastReason = null;

		resetModule();
	}

	/**
	 * Resets the current simulation module.
	 * 
	 * @since 0.2.2
	 */
	public abstract void resetModule();

	/**
	 * This method is used during development to check algorithms for the online simulator.
	 * 
	 * @param simKernel Simulation kernel
	 * @param netPlan Input network design
	 * @param eventGenerator Event generator
	 * @param customEventGeneratorParameters Custom event generator parameters (null means empty)
	 * @param eventProcessor Event processor
	 * @param customEventProcessorParameters Custom event processor parameters (null means empty)
	 * @param customSimulatorParameters Custom simulator parameters (null means empty)
	 * @param net2planParameters Net2Plan parameters
	 * @return Simulation report
	 * @since 0.3.0
	 */
	public static Pair<NetPlan, String> runSimulation(SimKernel simKernel, NetPlan netPlan, IExternal eventGenerator, Properties customEventGeneratorParameters, IExternal eventProcessor, Properties customEventProcessorParameters, Properties customSimulatorParameters, Map<String, String> net2planParameters)
	{
		simKernel.setNetPlan(netPlan);
		simKernel.checkLoadedNetPlan(simKernel.getInitialNetPlan());

		Map<String, String> eventGeneratorParameters = CommandLineParser.getParameters(eventGenerator.getParameters(), customEventGeneratorParameters);
		Map<String, String> eventProcessorParameters = CommandLineParser.getParameters(eventProcessor.getParameters(), customEventProcessorParameters);
		Map<String, String> simulationParameters = CommandLineParser.getParameters(simKernel.getSimulationParameters(), customSimulatorParameters);
		simKernel.configureSimulation(simulationParameters, net2planParameters, eventGenerator, eventGeneratorParameters, eventProcessor, eventProcessorParameters);
		simKernel.initialize();

		System.out.println("Net2Plan parameters");
		System.out.println("-----------------------------");
		System.out.println(StringUtils.mapToString(net2planParameters, "=", String.format("%n")));
		System.out.println();
		System.out.println("Simulation parameters");
		System.out.println("-----------------------------");
		System.out.println(StringUtils.mapToString(simulationParameters, "=", String.format("%n")));
		System.out.println();
		System.out.println(simKernel.getEventGeneratorLabel() + " parameters");
		System.out.println("-----------------------------");
		System.out.println(eventGeneratorParameters.isEmpty() ? "None" : StringUtils.mapToString(eventGeneratorParameters, "=", String.format("%n")));
		System.out.println();
		System.out.println(simKernel.getEventProcessorLabel() + " parameters");
		System.out.println("-----------------------------");
		System.out.println(eventProcessorParameters.isEmpty() ? "None" : StringUtils.mapToString(eventProcessorParameters, "=", String.format("%n")));
		System.out.println();
		System.out.println("Simulation started...");
		System.out.println();

		long init = System.nanoTime();
		simKernel.getSimCore().setSimulationState(SimCore.SimState.RUNNING);
		simKernel.getSimCore().run();
		long end = System.nanoTime();
		if (simKernel.lastReason != null && !(simKernel.lastReason instanceof EndSimulationException)) throw new RuntimeException(simKernel.lastReason);
		
		double totalSimTimeInSeconds = (end - init) / 1.0e9;
		String totalSimTime = StringUtils.secondsToYearsDaysHoursMinutesSeconds(totalSimTimeInSeconds);
		System.out.println(String.format("%n%nSimulation finished successfully in %s", totalSimTime));
		
		return Pair.of(simKernel.getCurrentNetPlan(), simKernel.getSimulationReport());
	}
	
	/**
	 *
	 * @param stateListener
	 * @since 0.2.2
	 */
	public void setGUIListener(IGUISimulationListener stateListener)
	{
		if (this.guiListener != null) throw new RuntimeException("A state listener was already installed");
		this.guiListener = stateListener;
	}

	/**
	 * <p>Sets the initial network plan.</p>
	 * 
	 * <p><b>Important</b>: Once the simulation is started, the initial network
	 * plan cannot be changed.</p>
	 * 
	 * @param netPlan Initial network plan
	 * @since 0.2.2
	 */
	public void setNetPlan(NetPlan netPlan)
	{
		if (simCore.getSimulationState() != SimCore.SimState.NOT_STARTED)
			throw new Net2PlanException("Network design cannot be changed once simulation was started");
	
		checkLoadedNetPlan(netPlan);
		originalNetPlan = null;
		initialNetPlan = netPlan.unmodifiableView();
		initializeNetState();
	}

	/**
	 * Processes the event in the simulation. In case that new events are to 
	 * be generated, they should be scheduled calling to 
	 * {@link #scheduleEvent(com.net2plan.interfaces.simulation.SimEvent) scheduleEvent}.
	 * 
	 * @param event Current event
	 * @since 0.3.0
	 */
	public abstract void simulationLoop(SimEvent event);
}